# !/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Request Rate Limiter initializer"""

from slowapi import Limiter
from slowapi.util import get_remote_address

API_RATE_LIMITER = Limiter(key_func=get_remote_address, default_limits=["5/minute"])



#!/usr/bin/env python

#  SPDX-FileCopyrightText: 2019-2025 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""This is the module that contains the tools for the CDS API.

The cdsapi package and methods used to interact with the CDS API are defined here.
These are accredited to the Copernicus Climate Data Store (CDS) and the European Centre for Medium-Range Weather
Forecasts (ECMWF) and are licensed under the Apache License, Version 2.0 which can be found at
https://www.apache.org/licenses/LICENSE-2.0.

"""

from datetime import date
from enum import Enum

import cdsapi
from loguru import logger
from pydantic import BaseModel, Field

_DEFAULT_VARIABLES = [
    "stl1",
    "stl2",
    "stl3",
    "stl4",
    "swvl1",
    "swvl2",
    "swvl3",
    "swvl4",
]


def _info_callback(*args, **kwargs) -> None:  # noqa: ANN002, ANN003
    """This is a callback function that is used to print information about the download process."""
    if len(args) > 0 or len(kwargs) > 0:
        logger.info("Callback received:")
        logger.info(" - args: ", *args)
        logger.info(" - kwargs: ", **kwargs)


class CDSDataSets(str, Enum):
    """Currently supported datasets for the CDS API."""

    ERA5SL = "reanalysis-era5-single-levels"
    ERA5LAND = "reanalysis-era5-land"


class CDSRequest(BaseModel):
    """A class that holds all necessary information for a CDS API request."""

    product_type: list[str] = Field(["reanalysis"])
    variables: list[str]
    year: list[str] = Field([date.strftime(date.today(), "%Y")])
    month: list[str] = Field([date.strftime(date.today(), "%m")])
    day: list[str] = Field([date.strftime(date.today(), "%d")])
    time: list[str] = Field(
        [
            "00:00",
            "01:00",
            "02:00",
            "03:00",
            "04:00",
            "05:00",
            "06:00",
            "07:00",
            "08:00",
            "09:00",
            "10:00",
            "11:00",
            "12:00",
            "13:00",
            "14:00",
            "15:00",
            "16:00",
            "17:00",
            "18:00",
            "19:00",
            "20:00",
            "21:00",
            "22:00",
            "23:00",
        ],
    )
    data_format: str = "netcdf"
    download_format: str = "zip"
    area: tuple[float, float, float, float] = (53.7, 3.2, 50.75, 7.22)

    @property
    def request_parameters(self) -> dict[str, str | list[str] | tuple[float]]:
        """Returns the request parameters as a dictionary."""
        return {
            "product_type": self.product_type,
            "format": self.data_format,
            "variable": self.variables,
            "year": self.year,
            "month": self.month,
            "day": self.day,
            "time": self.time,
            "area": self.area,
            "data_format": self.data_format,
            "download_format": self.download_format
        }


CDS_CLIENT = cdsapi.Client(info_callback=_info_callback(), url="https://cds.climate.copernicus.eu/api")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""KNMI current weather data aggregate fetcher."""

import copy
from datetime import datetime
from typing import List, Optional

import numpy as np
import xarray as xr
from dateutil.relativedelta import relativedelta

from weather_provider_api.routers.weather.base_models.model import WeatherModelBase
from weather_provider_api.routers.weather.sources.knmi.client.actuele_waarnemingen_register_repository import (
    ActueleWaarnemingenRegisterRepository,
)
from weather_provider_api.routers.weather.sources.knmi.stations import stations_actual
from weather_provider_api.routers.weather.sources.knmi.utils import (
    find_closest_stn_list,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition
from weather_provider_api.routers.weather.utils.pandas_helpers import coords_to_pd_index


class ActueleWaarnemingenRegisterModel(WeatherModelBase):
    """A Weather model aimed at accessing a 24-hour register for the "KNMi Actuele Waarnemingen" dataset."""

    def is_async(self):
        return self.async_model

    def __init__(self):
        super().__init__()
        self.id = "waarnemingen_register"
        self.name = "KNMI Actuele Waarnemingen - 48 uur register"
        self.version = ""
        self.url = "https://www.knmi.nl/nederland-nu/weer/waarnemingen"
        self.predictive = False
        self.time_step_size_minutes = 10
        self.num_time_steps = 12
        self.description = "48 Hour register for current weather observations. Updated every 10 minutes."
        self.async_model = False
        self.repository = ActueleWaarnemingenRegisterRepository()

        self.to_si = {
            "weather_description": {
                "name": "weather_description",
                "convert": self.no_conversion,
            },
            "temperature": {"convert": self.celsius_to_kelvin},
            "humidity": {"convert": self.percentage_to_frac},
            "wind_direction": {"convert": self.dutch_wind_direction_to_degrees},
            "wind_speed": {"convert": self.no_conversion},  # m/s
            "visibility": {"convert": self.no_conversion},  # m
            "air_pressure": {"convert": lambda x: x * 100},  # hPa to Pa
        }
        self.to_human = copy.deepcopy(self.to_si)
        self.to_human["temperature"]["convert"] = self.no_conversion  # C

        self.human_to_model_specific = self._create_reverse_lookup(self.to_si)

    def get_weather(
        self,
        coords: List[GeoPosition],
        begin: Optional[np.datetime64],
        end: Optional[np.datetime64],
        weather_factors: List[str] = None,
    ) -> xr.Dataset:
        """The function that gathers and processes the requested Actuele Waarnemingen Register weather data from the
        48-hour register and returns it as a Xarray Dataset.
        (The register for this model interprets directly from an HTML page, but the information is also available from
        the data platform. Due to it being rather impractically handled, we stick to the site for now.)

        Args:
            coords:             A list of GeoPositions containing the locations the data is requested for.
            begin:              A datetime containing the start of the period to request data for.
            end:                A datetime containing the end of the period to request data for.
            weather_factors:    A list of weather factors to request data for (in string format)

        Returns:
            An Xarray Dataset containing the weather data for the requested period, locations and factors.

        Notes:
            As this model only return the current weather data the 'begin' and 'end' values are not actually used.
        """
        updated_weather_factors = self._request_weather_factors(weather_factors)
        coords_stn, _, _ = find_closest_stn_list(stations_actual, coords)

        now = datetime.utcnow()
        if now - relativedelta(days=1) > begin:
            raw_ds = self.repository.get_48_hour_registry_for_station(station=coords_stn)
        else:
            raw_ds = self.repository.get_24_hour_registry_for_station(station=coords_stn)

        data_dictionary = {
            var_name: (["time", "coord"], var.values)
            for var_name, var in raw_ds.data_vars.items()
            if var_name in updated_weather_factors and var_name not in ["lat", "lon"]
        }

        timeline = raw_ds.coords["time"].values

        output_ds = xr.Dataset(
            data_vars=data_dictionary,
            coords={"time": timeline, "coord": coords_to_pd_index(coords)},
        )
        output_ds = output_ds.unstack("coord")
        return output_ds

    def _request_weather_factors(self, factors: Optional[List[str]]) -> List[str]:
        # Implementation of the Base Weather Model function that returns a list of known weather factors for the model.
        if factors is None:
            return list(self.to_si.keys())

        new_factors = []

        for f in factors:
            f_low = f.lower()
            if f_low in self.to_si:
                new_factors.append(f_low)

        return list(set(new_factors))  # Cleanup any duplicate values and return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""This module houses the repository class for the Actuele Waarnemingen Register."""

from datetime import datetime
from typing import List

import xarray as xr
from dateutil import tz
from dateutil.relativedelta import relativedelta
from loguru import logger

from weather_provider_api.routers.weather.repository.repository import (
    WeatherRepositoryBase,
)
from weather_provider_api.routers.weather.sources.knmi.utils import (
    download_actuele_waarnemingen_weather,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition
from weather_provider_api.routers.weather.utils.grid_helpers import (
    round_coordinates_to_wgs84_grid,
)


class ActueleWaarnemingenRegisterRepository(WeatherRepositoryBase):
    """ """

    def _delete_files_outside_of_scope(self):
        logger.info(f"Deleting files outside of scope [{self.first_day_of_repo} - {self.last_day_of_repo}]")
        if self.filename.exists():
            current_data = xr.load_dataset(self.filename, engine="netcdf4")
            current_data = current_data.sel(time=slice(self.first_day_of_repo, self.last_day_of_repo))
            current_data.to_netcdf(self.filename, format="NETCDF4")

    def _get_file_list_for_period(self, start: datetime, end: datetime):
        return self.filename

    def __init__(self):
        # Pre-work
        super().__init__()

        self.repository_name = "KNMI Actuele Waarnemingen - 48 uur register"

        # Repository settings
        self.file_prefix = "ACTUEEL48"
        self.runtime_limit = 3 * 60  # Three minutes maximum runtime
        self.permanent_suffixes = []
        self.filename = self.repository_folder / f"{self.file_prefix}_register.nc"
        self.file_identifier_length = 8

        logger.debug(f"Initialized the [{self.repository_name}] repository")

    @property
    def repository_sub_folder(self):
        return "ACTUEEL48"  # not using self.file_prefix, because that may not exist the first time using this

    def _get_repo_sub_folder(self):
        return self.repository_sub_folder

    @property
    def first_day_of_repo(self) -> datetime:
        # Property to get the first moment of the repository as translated to the Dutch timezone.
        from_zone = tz.gettz("UTC")
        to_zone = tz.gettz("Europe/Amsterdam")

        first_day_of_repo = datetime.utcnow() - relativedelta(hours=48)  # 48 hours
        first_day_of_repo = first_day_of_repo.replace(tzinfo=from_zone)
        first_day_of_repo = first_day_of_repo.astimezone(to_zone)
        return first_day_of_repo.replace(tzinfo=None)

    @property
    def last_day_of_repo(self) -> datetime:
        # Property to get the last moment of the repository as translated to the Dutch timezone.
        from_zone = tz.gettz("UTC")
        to_zone = tz.gettz("Europe/Amsterdam")

        last_day_of_repo = datetime.utcnow()  # Right now
        last_day_of_repo = last_day_of_repo.replace(tzinfo=from_zone)
        last_day_of_repo = last_day_of_repo.astimezone(to_zone)

        return last_day_of_repo.replace(tzinfo=None)

    def update(self):
        """Implementation of the WeatherRepository update method

        Attempts to update the repository with the current data. For this repository, that means getting the current
         Actuele Waarnemingen output and storing it, while removing any data not within the scope of the repository.

        Returns:
            Nothing.

        """
        raw_weather_ds = download_actuele_waarnemingen_weather()
        time = datetime.utcnow().replace(second=0, microsecond=0)

        # Cleanup any old data
        self.cleanup()

        # Update the file
        self._update_file_with_new_data(new_data_ds=raw_weather_ds, update_moment=time)

    def _update_file_with_new_data(self, new_data_ds: xr.Dataset, update_moment: datetime):
        """This method updates any existing data file with new data or creates a new from scratch if needed, using the
         given dataset.

        Args:
            new_data_ds (xr.Dataset):   A Xarray Dataset holding the data to append / create the data file with.
            update_moment (datetime):   A datetime holding the moment of update. Note that this doesn't need to be the
                                         moment that the data is from.

        Returns:
            Nothing. The file is just updated.

        """
        # Opening the file:
        if self.filename.exists():
            stored_data_ds = xr.load_dataset(self.filename, engine="netcdf4")
        else:
            stored_data_ds = None

        logger.info(
            f"Storing data at [{update_moment.strftime('%m-%d-%Y %H:%M:%S')}] for "
            f"[{new_data_ds.isel(STN=0, time=0)['time'].values}]"
        )
        if stored_data_ds is None:
            new_data_ds.to_netcdf(self.filename, format="NETCDF4")
        else:
            # Check if time not already in system
            try:
                new_stored_data_ds = xr.merge([new_data_ds, stored_data_ds])

                new_stored_data_ds.to_netcdf(self.filename, format="NETCDF4")
            except ValueError as value_error:
                logger.error(f"Could not update file: {value_error}")

    def get_24_hour_registry_for_station(self, station: int) -> xr.Dataset:
        """This method obtains the last 24 hours of data of Actuele Waarnemingen and returns it for single station.

        Args:
            station (int):  An integer representing the station to gather data for

        Returns:
            xr.Dataset: A Xarray Dataset holding last 24 hours of data for the requested station

        """
        stored_data_ds = xr.load_dataset(self.filename, engine="netcdf4")
        return stored_data_ds.sel(
            STN=station,
            time=slice(self.first_day_of_repo + relativedelta(days=1), self.last_day_of_repo),
        )

    def get_48_hour_registry_for_station(self, station: int) -> xr.Dataset:
        """This method obtains the last 48 hours of data of Actuele Waarnemingen and returns it for single station.

        Args:
            station (int):  An integer representing the station to gather data for

        Returns:
            xr.Dataset: A Xarray Dataset holding last 48 hours of data for the requested station

        """
        stored_data_ds = xr.load_dataset(self.filename, engine="netcdf4")
        stored_data_ds = stored_data_ds.sel(
            STN=station,
            time=slice(self.first_day_of_repo, self.last_day_of_repo),
        )
        return stored_data_ds

    def get_grid_coordinates(self, coordinates: List[GeoPosition]) -> List[GeoPosition]:
        """Rounds a list of GeoPositions to the resolution set through grid_resolution"""
        return round_coordinates_to_wgs84_grid(coordinates, (0.023, 0.037), (49, 0))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import tempfile
from typing import List, Tuple, Union

import pandas as pd
import xarray
from starlette.responses import FileResponse

from weather_provider_api.routers.weather.api_models import (
    ResponseFormat,
    ScientificJSONResponse,
    WeatherContentRequestMultiLocationQuery,
    WeatherContentRequestQuery,
)


def file_or_text_response(
    unserialized_data: xarray.Dataset,
    response_format: ResponseFormat,
    source_id: str,
    model_id: str,
    request: Union[WeatherContentRequestQuery, WeatherContentRequestMultiLocationQuery],
    coords: List[Tuple[float, float]],
) -> tuple[ScientificJSONResponse | FileResponse, str | None]:
    if response_format == ResponseFormat.json:
        return json_response(unserialized_data, coords)
    elif response_format == ResponseFormat.json_dataset:
        return json_dataset_response(unserialized_data)
    else:
        return file_response(unserialized_data, response_format, source_id, model_id, request, coords)


def file_response(
    unserialized_data: xarray.Dataset,
    response_format: ResponseFormat,
    source_id: str,
    model_id: str,
    request: WeatherContentRequestQuery,
    coords: list[tuple[float, float]],
) -> tuple[FileResponse, str]:
    if response_format == ResponseFormat.netcdf4:
        file_path = to_netcdf4(unserialized_data)
        mime = "application/x-netcdf4"
        extension = ".v4.nc"
    elif response_format == ResponseFormat.netcdf3:
        file_path = to_netcdf3(unserialized_data)
        mime = "application/x-netcdf3"
        extension = ".v3.nc"
    elif response_format == ResponseFormat.csv:
        file_path = to_csv(unserialized_data, coords)
        mime = "text/csv"
        extension = ".csv"
    else:
        raise NotImplementedError(f"Cannot create file response for the {response_format.name} response format")

    file_name = generate_filename(source_id, model_id, request, extension)
    response = FileResponse(file_path, media_type=mime, filename=file_name)
    return response, file_path


def generate_filename(source_id: str, model_id: str, request: WeatherContentRequestQuery, extension: str):
    file_name = f"weather_{source_id}_{model_id}_{request.begin}-{request.end}{extension}".replace(" ", "T").replace(
        ":", ""
    )
    return file_name


def to_netcdf4(unserialized_data: xarray.Dataset):
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    temp_file_path = temp_file.name
    unserialized_data.to_netcdf(path=temp_file_path, mode="w", format="NETCDF4", engine="netcdf4")
    return temp_file_path


def to_netcdf3(unserialized_data: xarray.Dataset):
    # SciPy handler used by Xarray for direct serialization to binary doesn't support v3. File-based one does.
    temp_file = tempfile.NamedTemporaryFile(delete=False)
    temp_file_path = temp_file.name
    unserialized_data.to_netcdf(path=temp_file_path, format="NETCDF3_64BIT", engine="netcdf4")
    return temp_file_path


def to_csv(unserialized_data: xarray.Dataset, coords):
    csv_strings = {}
    cols = []

    for i, c in enumerate(coords):
        df = get_weather_slice_for_coords(c, unserialized_data)

        if i == 0:
            cols = list(df.columns.values)
            cols.remove("lat")
            cols.remove("lon")
            cols = ["lat", "lon"] + cols

        csv_str = df.to_csv(columns=cols, header=False, float_format="%.4f")
        csv_strings[serialize_coords(c)] = csv_str

    cols = ["time"] + cols

    header = ",".join(cols)
    body = "\n".join(csv_strings.values())
    payload = (header + "\n" + body).encode("utf-8")

    temp_file = tempfile.NamedTemporaryFile(delete=False)
    temp_file.write(payload)
    temp_file.close()

    return temp_file.name


def get_weather_slice_for_coords(coord, unserialized_data) -> pd.DataFrame:
    # We use .sel to create slices from the unserialized dataset
    # We then switch to a Pandas dataframe
    if isinstance(coord, list):
        coord = coord[0]
    weather = unserialized_data.sel(lat=coord[0], lon=coord[1], method="nearest")
    if weather.sizes["time"] == 1:
        # Because a single moment in time can't be squeezed...
        df = weather.to_dataframe()
    else:
        # ... and multiple times need to be.
        df = weather.squeeze().to_dataframe()
    return df


def json_response(unserialized_data: xarray.Dataset, coords):
    serialized_data = []
    for coordinate in coords:
        df = get_weather_slice_for_coords(coordinate, unserialized_data)
        serialized_data.append(df.reset_index().to_dict(orient="records"))
    return ScientificJSONResponse(serialized_data), None


def json_dataset_response(unserialized_data: xarray.Dataset):
    data_dict = unserialized_data.to_dict()
    return ScientificJSONResponse(data_dict), None


def serialize_coords(coord):
    coord_str = ",".join([str(x) for x in coord])
    return coord_str



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from weather_provider_api.core.exceptions.api_models import ErrorResponseModel

ERROR_RESPONSE = {"model": ErrorResponseModel}



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""API Validation checks"""

import re
from datetime import datetime

from fastapi import FastAPI
from loguru import logger
from starlette.middleware.base import BaseHTTPMiddleware
from starlette.requests import Request

from weather_provider_api.config import APP_CONFIG
from weather_provider_api.core.exceptions.exceptions import APIExpiredException
from weather_provider_api.core.initializers.exception_handling import (
    handle_http_exception,
)


def initialize_api_validation(application: FastAPI):
    """Method for attach the API validity checker to a FastAPI application.

    Args:
        application:    The FastAPI application to attach the checker to.

    Returns:
        Nothing. The application itself is updated.

    """

    async def check_api_for_validity(request: Request, call_next):
        """The method that validated the API validity.

        Args:
            request:    The request to evaluate
            call_next:  The call_next object for the request (what to do next if this step doesn't raise any exceptions)

        Returns:
            The next step to execute for this request. This is either the original call_next, or an
             HTTP Exception trigger for an APIExpiredException.

        """
        api_prefix = r"/api/v"
        api_suffix = "/weather"
        request_url = str(request.url)
        today = datetime.today().date()
        response = None

        api_version_in_url = re.search(f"{api_prefix}\\d+{api_suffix}", request_url)  # Looks for mentioning of version

        if not api_version_in_url:
            return await call_next(request)  # No API call, no problem

        # Validate the base application
        base_expiration_date = datetime.strptime(APP_CONFIG["base"]["expiration_date"], "%Y-%m-%d").date()
        if today > base_expiration_date:
            response = await handle_http_exception(
                request,
                APIExpiredException(
                    f"The main project's expiry date of [{base_expiration_date}] has been reached. "
                    "Please contact the maintainer of this project!"
                ),
            )

        if not api_version_in_url:
            return await call_next(request)  # continue as normal if no api version is involved

        # Determine the API interpreter version used and its expiry date
        start_location_of_version_number_in_url = api_version_in_url.start() + len(api_prefix)
        end_location_of_version_number_in_url = api_version_in_url.end() - len(api_suffix)

        api_version_used = request_url[start_location_of_version_number_in_url:end_location_of_version_number_in_url]

        # Validate the expiry date of the specific API version used
        version_expiration_date = datetime.strptime(
            APP_CONFIG[f"api_v{api_version_used}"]["expiration_date"], "%Y-%m-%d"
        ).date()
        if today > version_expiration_date:
            if not response:
                response = await handle_http_exception(
                    request,
                    APIExpiredException(
                        f"The expiry date for the v{api_version_used} interface of this API has been reached: "
                        f"[{version_expiration_date}] Please contact the maintainer of this project!"
                    ),
                )
            else:
                response = await handle_http_exception(
                    request,
                    APIExpiredException(
                        f"The expiry dates for both the main project [{base_expiration_date}] and the "
                        f"v{api_version_used} interface [{version_expiration_date}] of this API have been reached."
                        "Please contact the maintainer of this project!"
                    ),
                )

        # Handle the request normally if no expiry was triggered
        if not response:
            response = await call_next(request)

        return response

    application.add_middleware(BaseHTTPMiddleware, dispatch=check_api_for_validity)
    logger.info("Attached API Expiration Checking to the application...")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""-- MODULE --"""

import copy
from datetime import datetime, time, timedelta
from typing import List, Union

import xarray as xr
from loguru import logger

from weather_provider_api.routers.weather.base_models.model import WeatherModelBase
from weather_provider_api.routers.weather.sources.knmi.client.arome_repository import (
    HarmonieAromeRepository,
)
from weather_provider_api.routers.weather.sources.knmi.knmi_factors import arome_factors
from weather_provider_api.routers.weather.utils.date_helpers import (
    validate_begin_and_end,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


class HarmonieAromeModel(WeatherModelBase):
    """A weather model that incorporates the 'KNMI - Harmonie Arome' weather dataset into the Weather Provider
     Libraries and API.

    'KNMI - Harmonie Arome' is a predictive model for the upcoming 48 hours that gets generated every 6 hours.
    """

    def __init__(self):
        # Pre-work
        super().__init__()
        self.id = "arome"
        logger.debug(f"Initializing weather model: {self.id}")

        # Setting the model
        self.repository = HarmonieAromeRepository()
        self.name = "harmonie_arome_cy_p1"
        self.version = "0.3"
        self.url = "ftp://data.knmi.nl/download/harmonie_arome_cy40_p1/0.2/"
        self.predictive = True
        self.time_step_size_minutes = 60
        self.num_time_steps = 48
        self.async_model = False
        self.dataset_name = "harmonie_arome_cy40_p1"
        self.dataset_version = "0.2"

        # Put up conversion settings
        si_conversion_dict = {
            k: {"name": k, "convert": lambda x: x} for k in arome_factors.values()
        }  # The default output format for Arome is SI
        self.to_si = si_conversion_dict

        # Human output conversion:
        self.to_human = copy.deepcopy(self.to_si)
        self.to_human["temperature"]["convert"] = self.kelvin_to_celsius

        # Initialization complete
        logger.debug(f'The Weather model "{self.id}" was successfully initialized')

    def get_weather(
        self,
        coords: List[GeoPosition],
        begin: datetime = None,
        end: datetime = None,
        weather_factors: List[str] = None,
    ) -> xr.Dataset:
        """Implementation of the WeatherModelBase get_weather function that fetches weather data and returns it as a
        Xarray DataSet.

        """
        # Handle possibly missing values for this model
        begin = begin or datetime.combine(datetime.today(), time.min)  # Fallback: start of today
        end = end or datetime.combine(datetime.today() + timedelta(days=1), time.max)  # Fallback: end of tomorrow

        # Validate the given timeframe
        valid_begin, valid_end = validate_begin_and_end(
            begin,
            end,
            self.repository.first_day_of_repo,
            self.repository.last_day_of_repo,
        )

        # Gather the period from the repository
        weather_dataset = self.repository.gather_period(begin=valid_begin, end=valid_end, coordinates=coords)

        # Filter the requested factors
        translated_factors = []
        if weather_factors is not None:
            for factor in weather_factors:
                if factor in arome_factors.keys():
                    new_factor = arome_factors[factor]
                elif factor in arome_factors.values():
                    new_factor = factor
                else:
                    new_factor = None

                if new_factor is not None:
                    translated_factors.append(new_factor)

            for factor in weather_dataset.keys():
                if not any(item in factor for item in translated_factors):
                    weather_dataset = weather_dataset.drop(factor)

        return weather_dataset

    def is_async(self):
        return self.async_model

    def _request_weather_factors(self, factor_list: Union[List[str], None]) -> List[str]:
        """Implementation of the Weather Model Base function that returns a list of the known weather factors out of
        a given list for this model.

        """
        if not factor_list:
            return list(self.to_si.keys())

        validated_factors = []

        for factor in factor_list:
            factor_lowercase = factor.lower()

            if factor_lowercase in self.to_si:
                validated_factors.append(factor_lowercase)
            elif factor in self.to_si:
                validated_factors.append(factor)

        return list(set(validated_factors))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Class to retrieve the current Weather Alert status according to the KNMI site"""

from enum import Enum

import requests
from bs4 import BeautifulSoup
from requests.adapters import HTTPAdapter
from requests.exceptions import ProxyError, Timeout, TooManyRedirects
from urllib3 import Retry


class WeatherAlertCode(Enum):
    # Enum class with valid Weather Alert Codes
    green = "green"
    yellow = "yellow"
    orange = "orange"
    red = "red"


class WeatherAlert:
    """A class (not a Weather Model!) that parses the Weather Alert status from the KNMI site (Weeralarm)"""

    def __init__(self):
        self.id = "weatheralert"
        self.name = "KNMI Weather Alert"
        self.version = "0.8"
        self.url = "https://www.knmi.nl/nederland-nu/weer/waarschuwingen/"
        self.predictive = True
        self.provinces = (
            "drenthe",
            "friesland",
            "gelderland",
            "groningen",
            "flevoland",
            "limburg",
            "noord-brabant",
            "noord-holland",
            "overijssel",
            "utrecht",
            "zeeland",
            "zuid-holland",
        )  # The Dutch Provinces. Every province has its own page.

    def get_alarm(self):
        """A function that retrieves the current weather alarm stage for each of the Dutch provinces and puts those
            together into a formatted list of results (string-based).

        Returns:
            A list of strings holding all the provinces and their retrieved current alarm stages according to KNMI
        """
        alarm_list = []
        for province in self.provinces:
            # Every province is available from a different page, so we have to request all of them separately
            page_text = ""
            try:
                page = self._requests_retry_session().get(self.url + province)
                status_code = page.status_code
                page_text = page.text
            except Timeout:
                status_code = 408
            except ProxyError:
                status_code = 407
            except TooManyRedirects:
                status_code = 999

            append_string = self.process_page(page_text, status_code, province)
            alarm_list.append(append_string)
        return alarm_list

    @staticmethod
    def process_page(page_text: str, status_code, province):
        """A function that parses the weather alert page for a province and retrieves its current alarm stage.

        Args:
            page_text:      The response content retrieved while trying to download the page
            status_code:    The status code retrieved while trying to download the page.
            province:       The province associated with the url, status code and alarm stage.

        Returns:
            A tuple holding the province and a result-string for that province.
            A result-string usually hold the alarm stage for that province, but can also hold exceptions when
            downloading did not succeed as intended.
        """
        if status_code == 200 and page_text is not None:
            # A page was found and loaded
            soup = BeautifulSoup(page_text, features="lxml")

            classes_first_warning_block = soup.find("div", {"class": "warning-overview"})
            if classes_first_warning_block is not None:
                classes_first_warning_block = classes_first_warning_block["class"]

                for class_name in classes_first_warning_block:
                    if len(class_name) > len("warning-overview") and class_name[len("warning-overview--") :] in set(
                        item.value for item in WeatherAlertCode
                    ):
                        return province, class_name[len("warning-overview--") :]

            # If no valid code was found return an invalid data message
            return province, "could not find expected data on page"
        elif status_code == 408:
            return province, "time out op loading page"
        elif status_code == 407:
            return province, "proxy error on loading page"
        else:
            return province, "page proved inaccessible"

    @staticmethod
    def _requests_retry_session(
        # A function for basic retrying of an url when it isn't accessible immediately.
        retries=8,
        backoff_factor=0.01,
        status_forcelist=(500, 502, 504),
        session=None,
    ) -> requests.Session:
        session = session or requests.Session()
        retry = Retry(
            total=retries,
            read=retries,
            connect=retries,
            backoff_factor=backoff_factor,
            status_forcelist=status_forcelist,
        )
        adapter = HTTPAdapter(max_retries=retry)
        session.mount("https://", adapter)
        return session



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Version detection module."""

from importlib import metadata

import tomli
from loguru import logger


def _get_app_version() -> str:
    """This method tries to identify the API's main version identifier and return it.

    Returns:
        str:    The found version identifier, if found. If no version identifier was found, a warning value is returned
                 instead.

    """
    # First attempt: Get the version number by looking for a pyproject.toml file in the working directory.
    # Please note that this assumes that this function was called from a context that has the Project's main folder as
    #  the working directory.
    try:
        with open("./pyproject.toml", mode="rb") as project_file:
            version = tomli.load(project_file)["tool"]["poetry"]["version"]
        logger.info(f"Retrieved the project version from the pyproject.toml file: {version}")
        return version
    except FileNotFoundError as fnf_error:
        logger.debug(f"Could not retrieve the active version from the pyproject.toml file: {fnf_error}")

    # Second attempt: Get the version number from the package that was used to install this component, if applicable.
    try:
        version = metadata.version(__package__)
        logger.info(f"Retrieved the project version from package data: {version}")
        return version
    except metadata.PackageNotFoundError as pnf_error:
        logger.debug(f"Could not retrieve the active version from package data: {pnf_error}")

    # No version could be determined
    logger.warning("No version could be found for the project!")
    return "<< version could not be determined >>"


APP_VERSION = _get_app_version()



#!/usr/bin/env python

#  SPDX-FileCopyrightText: 2019-2025 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0
import glob
import tempfile
import zipfile
from datetime import date, datetime
from enum import Enum
from pathlib import Path

import xarray as xr
from dateutil.relativedelta import relativedelta
from loguru import logger
from pydantic import BaseModel
from pytz import UTC

from weather_provider_api.routers.weather.repository.repository import RepositoryUpdateResult
from weather_provider_api.routers.weather.sources.cds.client.cds_api_tools import CDS_CLIENT, CDSDataSets, CDSRequest


class Era5FileSuffixes(str, Enum):
    """Enum class for the different suffixes that can be used for the ERA5 files."""

    FORMATTED = ".FORMATTED.nc"
    UNFORMATTED = ".UNFORMATTED.nc"
    INCOMPLETE = ".INCOMPLETE.nc"
    TEMP = ".TEMP.nc"


class Era5UpdateSettings(BaseModel):
    """A class that holds the settings for updating the ERA5 data."""

    era5_dataset_to_update_from: CDSDataSets
    era5_product_type: str = "reanalysis"
    filename_prefix: str
    target_storage_location: Path
    repository_time_range: tuple[datetime, datetime]
    factors_to_process: list[str]
    factor_dictionary: dict[str, str]
    maximum_runtime_in_minutes: int = 2 * 60  # 2 hours


def era5_repository_update(update_settings: Era5UpdateSettings, test_mode: bool) -> RepositoryUpdateResult:
    """A function to update a variant of ERA5 data into the repository."""
    starting_moment_of_update = datetime.now(UTC)
    cutoff_time = starting_moment_of_update + relativedelta(minutes=update_settings.maximum_runtime_in_minutes)
    logger.info(
        f"Starting update of ERA5 data for {update_settings.era5_dataset_to_update_from} "
        f"to: {update_settings.target_storage_location}"
    )
    logger.debug(f" - Attempting update for time range: {update_settings.repository_time_range}")
    logger.debug(f" - Factors to process: {update_settings.factors_to_process}")
    logger.info(f" - Maximum runtime: {update_settings.maximum_runtime_in_minutes} minutes ({cutoff_time})")

    try:
        _era5_update_month_by_month(update_settings, starting_moment_of_update, cutoff_time, test_mode)
    except Exception as e:
        logger.error(f"Failed to update ERA5 data. Reason: {e}")
        return RepositoryUpdateResult.failure

    ending_moment_of_update = datetime.now(UTC)
    logger.info(
        f"Update finished. Total runtime: {(ending_moment_of_update - starting_moment_of_update).total_seconds()}"
    )
    return RepositoryUpdateResult.completed


def _era5_update_month_by_month(
    update_settings: Era5UpdateSettings, starting_moment_of_update: datetime, cutoff_time: datetime, test_mode: bool
):
    """A function to update a variant of ERA5 data into the repository."""
    amount_of_months_processed = amount_of_months_not_processable = 0
    average_time_per_month_in_minutes = 35

    update_month = _get_update_month(update_settings)
    target_update_month = update_settings.repository_time_range[0].replace(
        day=1, hour=0, minute=0, second=0, microsecond=0
    )

    while update_month > target_update_month:
        logger.info(f" > Processing month: {update_month.year}-{update_month.month}")
        if datetime.now(UTC) + relativedelta(minutes=average_time_per_month_in_minutes) > cutoff_time:
            logger.warning(
                "MAXIMUM RUNTIME REACHED: ",
                cutoff_time,
                datetime.now(UTC) + relativedelta(minutes=average_time_per_month_in_minutes),
                average_time_per_month_in_minutes,
            )
            logger.warning("Maximum runtime reached. Stopping update.")
            break

        update_result = _era5_update_month(update_settings, update_month, test_mode)
        if update_result == RepositoryUpdateResult.failure:
            amount_of_months_not_processable += 1
        amount_of_months_processed += 1

        if amount_of_months_not_processable / amount_of_months_processed > 0.5:
            logger.warning("More than 50% of the months failed to process. Stopping update.")
            break

        average_time_per_month_in_minutes = (
            (datetime.now(UTC) - starting_moment_of_update).total_seconds() / 60 / amount_of_months_processed
        )

        update_month = update_month - relativedelta(months=1)

    logger.info(
        f"Processed {amount_of_months_processed} months, {amount_of_months_not_processable} months failed to process."
    )
    logger.info(f"Average time per month: {average_time_per_month_in_minutes} minutes")


def _era5_update_month(
    update_settings: Era5UpdateSettings, update_month: datetime, test_mode: bool
) -> RepositoryUpdateResult:
    """A function to update a variant of ERA5 data into the repository."""
    logger.debug(f" > Processing month: {update_month.year}-{update_month.month}")

    month_file_base = f"{update_settings.filename_prefix}_{update_month.year}_{update_month.month:02d}"
    month_file = update_settings.target_storage_location / f"{month_file_base}"
    threshold_date = (datetime.now(UTC) - relativedelta(days=5)).replace(day=1)

    if file_requires_update(month_file, update_month, threshold_date):
        logger.debug(f" > File {month_file} requires update.")
        month_file_name = month_file.with_suffix(Era5FileSuffixes.UNFORMATTED)

        # Only the first day of each month in test mode, otherwise all days:
        day = [str(i) for i in list(range(1, 32))] if not test_mode else ["1"]

        try:
            download_era5_data(
                update_settings.era5_dataset_to_update_from,
                CDSRequest(
                    product_type=[update_settings.era5_product_type],
                    variables=update_settings.factors_to_process,
                    year=[str(update_month.year)],
                    month=[str(update_month.month)],
                    day=day,
                    time=[f"{hour:02d}:00" for hour in range(24)],
                ),
                target_location=str(month_file_name),
            )

            logger.debug("Stored file at: ", month_file_name)

            _recombine_multiple_files(month_file_name)

            _format_downloaded_file(month_file_name, update_settings.factor_dictionary)

            month_file_name.rename(month_file.with_suffix(Era5FileSuffixes.FORMATTED))
            logger.debug("Renamed to: ", month_file.with_suffix(Era5FileSuffixes.FORMATTED))
            _finalize_formatted_file(month_file, update_month, threshold_date)

        except Exception as e:
            logger.error(f" > Failed to update ERA5 data for {update_month}. Reason: {e}")
            return RepositoryUpdateResult.failure

    return RepositoryUpdateResult.completed


def _get_update_month(update_settings: Era5UpdateSettings) -> datetime:
    NORMAL_FIRST_MOMENT_AVAILABLE_FOR_ERA5 = (datetime.now(UTC) - relativedelta(days=5)).replace(
        hour=0, minute=0, second=0, microsecond=0
    )
    update_moment = update_settings.repository_time_range[1]

    update_moment = (
        update_moment
        if update_moment < NORMAL_FIRST_MOMENT_AVAILABLE_FOR_ERA5
        else NORMAL_FIRST_MOMENT_AVAILABLE_FOR_ERA5
    )

    if update_moment == NORMAL_FIRST_MOMENT_AVAILABLE_FOR_ERA5:
        update_moment = _verify_first_day_available_for_era5(update_moment, update_settings)

    update_moment = update_moment.replace(day=1, hour=0, minute=0, second=0, microsecond=0)
    return update_moment


def _verify_first_day_available_for_era5(update_moment: datetime, update_settings: Era5UpdateSettings) -> datetime:
    """A function to verify the first day available for ERA5 data.

    Normally the first full day available for ERA5 data is the 5 days ago, however, due to the nature of the data
    it might be that data takes longer to fully process and be available. This function will verify if the data is
    available and if not, will return the correct first day available.
    """
    logger.debug(" > Verifying the first day available for ERA5 data.")

    while True:
        try:
            download_era5_data(
                dataset=update_settings.era5_dataset_to_update_from,
                cds_request=CDSRequest(
                    product_type=[update_settings.era5_product_type],
                    variables=["stl1"],  # A factor that exists in all supported ERA5 datasets
                    year=[str(update_moment.year)],
                    month=[str(update_moment.month)],
                    day=[str(update_moment.day)],
                    time=[f"{hour:02d}:00" for hour in range(2)],
                ),
                target_location=tempfile.NamedTemporaryFile().name,
            )
            break
        except Exception as e:
            logger.debug(f" > Failed to download ERA5 data for {update_moment}. Reason: {e}")
            update_moment = update_moment - relativedelta(days=1)

            if update_moment < update_settings.repository_time_range[1] - relativedelta(days=45):
                raise ValueError(
                    "The first day available for ERA5 data could not be found within 40 days of the target date. "
                    "Aborting update."
                )

    return update_moment


def _finalize_formatted_file(file_path: Path, current_moment: date, verification_date: date) -> None:
    """A function to finalize the formatted file."""
    incomplete_month = verification_date.replace(day=1)
    permanent_month = (verification_date - relativedelta(months=3)).replace(day=1)

    if not file_path.with_suffix(Era5FileSuffixes.FORMATTED).exists():
        logger.error(f"Formatted file {file_path} does not exist. Aborting finalization.")
        raise FileNotFoundError(f" > Formatted file {file_path} does not exist.")

    # Cleanup the old files
    for file_suffix in [Era5FileSuffixes.TEMP, Era5FileSuffixes.INCOMPLETE]:
        if file_path.with_suffix(file_suffix).exists():
            try:
                file_path.with_suffix(file_suffix).unlink()
            except Exception as e:
                logger.error(f" > Failed to remove temporary file {file_path.with_suffix(file_suffix)}: {e}")

    # Rename the file to its proper name:
    if current_moment == verification_date.replace(day=1):
        # Current month means an incomplete file
        file_path.with_suffix(Era5FileSuffixes.FORMATTED).rename(file_path.with_suffix(Era5FileSuffixes.INCOMPLETE))
        logger.debug(f"Month [{current_moment}] was renamed to: {file_path.with_suffix(Era5FileSuffixes.INCOMPLETE)}")
    elif permanent_month < current_moment < incomplete_month:
        # Non-permanent file
        file_path.with_suffix(Era5FileSuffixes.FORMATTED).rename(file_path.with_suffix(Era5FileSuffixes.TEMP))
        logger.debug(f"Month [{current_moment}] was renamed to: {file_path.with_suffix(Era5FileSuffixes.TEMP)}")
    else:
        # Permanent file
        file_path.with_suffix(Era5FileSuffixes.FORMATTED).rename(file_path.with_suffix(".nc"))
        logger.debug(f'Month [{current_moment}] was renamed to: {file_path.with_suffix(".nc")}')


def file_requires_update(file_path: Path, current_month: date, verification_date: date) -> bool:
    """A function that checks if a file requires an update based on the current state of the repository."""
    if file_path.with_suffix(".nc").exists():
        # A regular file exists, no updates required
        logger.debug(" > A regular file already exists: NO UPDATE REQUIRED")
        return False

    if file_path.with_suffix(Era5FileSuffixes.TEMP).exists():
        # If a file is temporary we only check for a permanent update if more than 3 months have past since the current
        # most recent date with data.
        threshold_date = (verification_date - relativedelta(months=3)).replace(day=1)
        if current_month < threshold_date:
            logger.debug(" > A temporary file exists within the update range: UPDATE REQUIRED")
            return True
        logger.debug(" > A temporary file exists within the update range: UPDATE REQUIRED")
        return False

    # A file exists but isn't any regular supported type to be updated
    if (
        file_path.with_suffix(Era5FileSuffixes.UNFORMATTED).exists()
        or file_path.with_suffix(Era5FileSuffixes.FORMATTED).exists()
    ):
        logger.debug(" > An unformatted file or formatted file exists: UPDATE REQUIRED")
        return True  # An update should both clean the UNFORMATTED file and generate a proper one

    if not file_path.with_suffix(".nc").exists() or file_path.with_suffix(Era5FileSuffixes.INCOMPLETE).exists():
        logger.debug(" > No file exists, or it is still incomplete: UPDATE REQUIRED")
        print("File path: ", file_path)
        return True  # No file matching the mask or incomplete files always mean the update is required!

    files_in_folder = glob.glob(f"{file_path}*.nc")
    logger.warning(
        f" > Unexpected files existed in the repository folder: {files_in_folder}. These should be dealt with."
    )
    return False


def _format_downloaded_file(unformatted_file: Path, allowed_factors: dict) -> None:
    """A function that formats the downloaded file to the correct format for the repository."""
    logger.info(f" > Formatting the downloaded file at: {unformatted_file}")
    ds_unformatted = load_file(unformatted_file)
    ds_unformatted.attrs = {}  # Remove unneeded attributes

    if "expver" in ds_unformatted.indexes.keys():
        # We remove the expver index used to denominate temporary data (5) and regular data (1) and add a field for it
        # NOTE: We removed the drop_sel version as it didn't quite have the same result as drop yet. Reverting until
        #  the proper use has been validated...
        ds_unformatted_expver5 = ds_unformatted.sel(expver=5).drop("expver").dropna("valid_time", how="all")
        ds_unformatted_expver1 = ds_unformatted.sel(expver=1).drop("expver").dropna("valid_time", how="all")

        # Recombine the data
        ds_unformatted = ds_unformatted_expver1.merge(ds_unformatted_expver5)
        ds_unformatted["is_permanent_data"] = False
    else:
        ds_unformatted["is_permanent_data"] = True

    # Rename the factors to their longer names:
    for factor in ds_unformatted.variables.keys():
        if factor in allowed_factors:
            ds_unformatted = ds_unformatted.rename_vars({factor: allowed_factors[factor]})

    # Rename and encode data where needed:
    ds_unformatted.valid_time.encoding["units"] = "hours since 2016-01-01"
    ds_unformatted = ds_unformatted.rename(name_dict={"latitude": "lat", "longitude": "lon", "valid_time": "time"})

    # Store the data
    ds_unformatted.to_netcdf(path=unformatted_file, format="NETCDF4", engine="netcdf4")


def load_file(file: Path) -> xr.Dataset:
    """A function that loads and returns the full data for a specific repository file as a Xarray Dataset.

    Args:
        file:   The filename (in the Path format by PathLib) specifying the file to load
    Returns:
        An Xarray Dataset containing all the weather data held within the specified file.

    """
    if file.exists():
        with xr.open_dataset(file) as ds:
            ds.load()
        return ds

    # Raise a FileNotFoundError if the file doesn't exist
    logger.error(f" > File [{file!s}] does not exist")
    raise FileNotFoundError


def _recombine_multiple_files(unformatted_file: Path) -> None:
    """A function that recombines multiple files into one."""
    logger.debug(f" > Recombining multiple files for: {unformatted_file}")

    # Create a temporary directory to store the files
    temp_dir = tempfile.mkdtemp()

    # Unpack the file
    with zipfile.ZipFile(unformatted_file, "r") as zip_ref:
        zip_ref.extractall(temp_dir)

    concatenated_dataset = xr.Dataset()
    files_to_load_in_order = [
        "data_stream-oper_stepType-instant",
        "data_stream-oper_stepType-accum",
        # TODO: Add the following file back in when we can properly handle it
        # "data_stream-wave_stepType-instant",  # Something about this data doesn't mesh well anymore with the rest...
    ]

    # TODO: Load, convert to dataframe, merge, convert back to xarray
    concatenated_dataset = xr.Dataset()
    for filename in files_to_load_in_order:
        file_path = Path(temp_dir).joinpath(f"{filename}.nc")
        if not file_path.exists():
            logger.error(f" > Required file {filename}.nc does not exist. Aborting recombination.")
            raise FileNotFoundError(f" > Required file {filename}.nc does not exist. Aborting recombination.")

        dataset = xr.open_dataset(file_path)
        dataset = dataset.drop("expver", errors="raise")

        if not concatenated_dataset.data_vars:
            concatenated_dataset = dataset.copy(deep=True)
        else:
            concatenated_dataset = xr.merge(
                [concatenated_dataset, dataset], join="outer", compat="no_conflicts", combine_attrs="override"
            )

    concatenated_dataset.to_netcdf(unformatted_file, format="NETCDF4", engine="netcdf4")
    # raise ValueError("This is not working yet")


def download_era5_data(
    dataset: CDSDataSets,
    cds_request: CDSRequest,
    target_location: str,
) -> None:
    """A function to download ERA5 data."""
    try:
        CDS_CLIENT.retrieve(
            dataset,
            cds_request.request_parameters,
            target_location,
        )

    except Exception as e:
        logger.error(f"Failed to download ERA5 data. Reason: {e}")
        raise e



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import os
import site
from pathlib import Path

from loguru import logger


async def remove_file(file_path):  # pragma: no cover
    if file_path is not None:
        try:
            file_to_rm = Path(file_path).resolve()
            logger.info("Removing temporary file", file_path=file_to_rm)
            if file_to_rm.exists() and file_to_rm.is_file():
                file_to_rm.unlink()
        except FileNotFoundError as e:
            logger.exception(e)
            raise
    return True


def get_var_map_file_location(filename: str) -> Path:
    var_map_folder = "var_maps"

    possible_main_folders = [
        Path(os.getcwd()),  # Running from main folder
        Path(os.getcwd()).parent,  # Running from weather_provider_api folder or scripts
        Path(site.getsitepackages()[-1]),  # Running as package
    ]

    for folder in possible_main_folders:
        possible_var_map_folder = folder.joinpath(var_map_folder)
        if possible_var_map_folder.exists():
            logger.info(f'"var_maps" folder was found at: {possible_var_map_folder}')
            return possible_var_map_folder.joinpath(filename)

    logger.exception(f"File was not found: {filename}")
    raise FileNotFoundError



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""TODO:
- Decouple weather factors from the get_weather function
- Add the async process for e.g. CDS and Harmonie
- Improve datetime slicing
- Fix datetime validation in Swagger UI on IE and Safari
- Improve output / error messages when no data is available
"""

from typing import List

import accept_types
from fastapi import APIRouter, BackgroundTasks, Depends, Header, HTTPException

from weather_provider_api.routers.weather.api_models import (
    WeatherContentRequestQuery,
    WeatherFormattingRequestQuery,
    WeatherModel,
    WeatherSource,
    result_mime_types,
)
from weather_provider_api.routers.weather.controller import WeatherController
from weather_provider_api.routers.weather.sources.weather_alert.weather_alert import (
    WeatherAlert,
)
from weather_provider_api.routers.weather.utils import serializers
from weather_provider_api.routers.weather.utils.date_helpers import parse_datetime
from weather_provider_api.routers.weather.utils.file_helpers import remove_file

app = APIRouter()

controller = WeatherController()


def header_accept_type(accept: str = Header(None)) -> str:
    # Matches the given string to the nearest result in result_mime_types and returns the matching value
    best_match_or_none = accept_types.get_best_match(accept, result_mime_types)
    return result_mime_types[best_match_or_none]


# @weather_provider_api.get("/sources", response_model=List[WeatherSource], tags=["sync", "async"])
@app.get("/sources", response_model=List[WeatherSource], tags=["sync"])
async def get_sources():  # pragma: no cover
    # List all the sources known by the controller
    return controller.get_sources()


# @weather_provider_api.get("/sources/{source_id}", response_model=WeatherSource, tags=["sync", "async"])
@app.get("/sources/{source_id}", response_model=WeatherSource, tags=["sync"])
async def get_source(source_id: str):  # pragma: no cover
    # List all the models supported by the selected source
    source_id = source_id.lower()
    return controller.get_source(source_id)


@app.get("/sources/{source_id}/models", response_model=List[WeatherModel], tags=["sync"])
async def get_sync_models(source_id: str):  # pragma: no cover
    # List all the synchronous models supported by the selected source
    source_id = source_id.lower()
    return controller.get_models(source_id, fetch_async=False)


@app.get("/sources/{source_id}/models/{model_id}", tags=["sync"])
async def get_sync_weather(
    source_id: str,
    model_id: str,
    cleanup_tasks: BackgroundTasks,
    ret_args: WeatherContentRequestQuery = Depends(),
    fmt_args: WeatherFormattingRequestQuery = Depends(),
    accept: str = Depends(header_accept_type),
):  # pragma: no cover
    """Function to gather data for a specific model using specific settings (location, period, factors, e.g.).
        The function then formats this data into the requested output format (file-format and selected output unit) and
        returns it.

    Args:
        source_id:  The identifier for the chosen source
        model_id:   The identifier for the chosen model
        cleanup_tasks: Holder for background tasks, in get_sync_weather solely used for file cleanup, hence the name
        ret_args:   Contains a WeatherContentRequestQuery item with all the settings to be used for data collection
        fmt_args:   Contains a WeatherFormattingRequestQuery item with all the setting to be used for formatting the
                    output
        accept:

    Returns:
        Returns a file in the requested output format, containing any weather data that matched the specifications from
        the WeatherContentRequestQuery.

    """
    # TODO: Append a proper definition of the accept arg
    source_id = source_id.lower()
    model_id = model_id.lower()
    coords = controller.lat_lon_to_coords(ret_args.lat, ret_args.lon)

    begin = parse_datetime(ret_args.begin, raise_errors=True, loc=["query", "begin"])
    end = parse_datetime(
        ret_args.end,
        round_missing_time_up=True,
        raise_errors=True,
        loc=["query", "end"],
    )

    try:
        weather_data = controller.get_weather(
            source_id,
            model_id,
            fetch_async=False,
            coords=coords,
            begin=begin,
            end=end,
            factors=ret_args.factors,
        )
    except FileNotFoundError as e:
        raise HTTPException(status_code=404, detail=e.args[0])

    if weather_data is None:
        raise HTTPException(status_code=404, detail="No data was found for the given period")

    response_format = fmt_args.response_format or accept

    converted_weather_data = controller.convert_names_and_units(
        source_id, model_id, False, weather_data, fmt_args.units
    )

    coords = [
        (lat_val, lon_val)
        for (lat_val, lon_val) in zip(
            converted_weather_data.coords["lat"].values,
            converted_weather_data.coords["lon"].values,
        )
    ]

    response, optional_file_path = serializers.file_or_text_response(
        converted_weather_data, response_format, source_id, model_id, ret_args, coords
    )
    cleanup_tasks.add_task(remove_file, optional_file_path)

    return response


# @weather_provider_api.get("/weeralarm")
@app.get("/alarms/knmi", tags=["alerts"])
async def get_alarm():  # pragma: no cover
    # Fetches the WeatherAlarm status for all the provinces from KNMI Weer Alarm and returns the results
    weather_alert = WeatherAlert()
    return weather_alert.get_alarm()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""CDS - ERA5 Single Levels Weatherdata Model"""

import copy
from datetime import datetime
from typing import List, Optional

import xarray
import xarray as xr
from loguru import logger

from weather_provider_api.routers.weather.base_models.model import WeatherModelBase
from weather_provider_api.routers.weather.sources.cds.client.era5land_repository import (
    ERA5LandRepository,
)
from weather_provider_api.routers.weather.sources.cds.factors import era5sl_factors
from weather_provider_api.routers.weather.utils.date_helpers import (
    validate_begin_and_end,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition


class ERA5LandModel(WeatherModelBase):
    """A Weather Model that incorporates the:
        ERA5 hourly data on single levels from 1979 to present
    dataset into the Weather Provider API
    """

    def __init__(self):
        super().__init__()
        self.id = "era5land"
        logger.debug(f"Initializing weather model [{self.id}]")

        self.name = "CDS: ERA5-Land"

        self.version = "0.3"
        self.url = "https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-land?tab=overview"
        self.predictive = False
        self.description = (
            "ERA5-Land is a reanalysis dataset providing a consistent view of the evolution of land variables over "
            "several decades at an enhanced resolution compared to ERA5. ERA5-Land has been produced by replaying the "
            "land component of the ECMWF ERA5 climate reanalysis. Reanalysis combines model data with observations "
            "from across the world into a globally complete and consistent dataset using the laws of physics. "
            "Reanalysis produces data that goes several decades back in time, providing an accurate description of "
            "the climate of the past."
        )
        self.async_model = False

        self.time_step_size_minutes = 60
        self.num_time_steps = 0

        self.repository = ERA5LandRepository()

        # Set up Conversion Dictionary
        si_conversion_dict = {
            k: {"name": k, "convert": lambda x: x} for k in era5sl_factors.values()
        }  # The default output format for ERA5SL is already SI
        self.to_si = si_conversion_dict

        # Human output conversion:
        self.to_human = copy.deepcopy(self.to_si)
        self.to_human["sea_surface_temperature"]["convert"] = self.kelvin_to_celsius
        self.to_human["soil_temperature_level_1"]["convert"] = self.kelvin_to_celsius
        self.to_human["soil_temperature_level_2"]["convert"] = self.kelvin_to_celsius
        self.to_human["soil_temperature_level_3"]["convert"] = self.kelvin_to_celsius
        self.to_human["soil_temperature_level_4"]["convert"] = self.kelvin_to_celsius
        self.to_human["2m_temperature"]["convert"] = self.kelvin_to_celsius

        logger.debug(f"Weather model [{self.id}] initialized successfully")

    def is_async(self):
        """Returns the async model status"""
        return self.async_model

    def get_weather(
        self,
        coords: List[GeoPosition],
        begin: datetime,
        end: datetime,
        weather_factors: List[str] = None,
    ) -> xr.Dataset:
        """The function that gathers and processes the requested ERA5 Single Levels weather data from the repository
            and returns it as a Xarray Dataset.

        Args:
            coords:             A list of GeoPositions containing the locations the data is requested for.
            begin:              A datetime containing the start of the period to request data for.
            end:                A datetime containing the end of the period to request data for.
            weather_factors:    A list of weather factors to request data for (in string format)

        Returns:
            An Xarray Dataset containing the weather data for the requested period, locations and factors.
        """
        # Test and account for invalid datetime timeframes or input
        begin, end = validate_begin_and_end(
            begin,
            end,
            self.repository.first_day_of_repo,
            self.repository.last_day_of_repo,
        )

        # Validate the requested weather factors:
        validated_factors = self._validate_weather_factors(weather_factors)

        ds = self._fill_dataset_with_data(coords, begin, end, validated_factors)
        return ds

    @staticmethod
    def _validate_weather_factors(weather_factors: List[str]) -> List[str]:
        """A function that validates a list of weather factors to that of the dataset in the repository.
            Existing factors will be kept, non-existing removed, and if the list is empty the full set for the dataset
            will be used.

        Args:
            weather_factors:    A list of weather factors to validate (in string format)

        Returns:
            A list of weather factors (in string format) only factors that match those of the ERA5SL dataset.
        """
        if weather_factors is None:
            weather_factors = [era5sl_factors[x] for x in list(era5sl_factors.keys())]

        # Lookup using the generic long name
        weather_factors_long_names = [x for x in weather_factors if x in era5sl_factors.values()]
        # Lookup using the CDS' own short name
        weather_factors_short_names = [era5sl_factors[x] for x in weather_factors if x in era5sl_factors.keys()]

        # Merge the results
        weather_factors = weather_factors_long_names + weather_factors_short_names

        # If nothing useful was found, just return everything
        return weather_factors

    @staticmethod
    def _get_list_of_factors_to_drop(factors: List[str], dataset_to_drop_from: xarray.Dataset) -> List[str]:
        # A small function that that compares a list of factors to keep with the full list, to make a list of factors
        # to drop from a full set.
        to_drop = [x for x in era5sl_factors.values() if x not in factors]
        logger.debug("Dropping the following factors for the request: " + str(to_drop))
        not_in_dataset = list(set(to_drop).difference(dataset_to_drop_from.keys()))
        if len(not_in_dataset) > 0:
            logger.warning(
                f"The following fields passed the factor validation, but weren't in the ERA5SL data: "
                f"{not_in_dataset}"
            )
        to_drop = [x for x in to_drop if x in dataset_to_drop_from.keys()]
        return to_drop

    def _fill_dataset_with_data(
        self,
        era5sl_coordinates: List[GeoPosition],
        begin: datetime,
        end: datetime,
        validated_factors: List[str],
    ) -> xr.Dataset:
        """A function that fills a dataset with ERA5SL weather data from the repository, based on the requested
            coordinates and period, and removes any not-requested weather factors from the output.

        Args:
            era5sl_coordinates:     A list of GeoPositions containing the locations to be gathered from the repository.
            begin:                  A datetime containing the starting moment for the period to gather.
            end:                    A datetime containing the end moment for the period to gather.
            validated_factors:      A list of valid ERA5SL factors wanted for the result.

        Returns:
            An Xarray Dataset containing the weather data requested.
        """
        # Gather a dataset with the proper period and coordinates
        ds = self.repository.gather_period(begin, end, era5sl_coordinates)
        ds = ds.sel(time=slice(begin, end))  # Slice of any overflowing time-range

        # Drop excess weather factors
        factors_to_drop = self._get_list_of_factors_to_drop(validated_factors, ds)
        ds = ds.drop_vars(factors_to_drop)
        return ds

    def _request_weather_factors(self, factors: Optional[List[str]]) -> List[str]:
        # Implementation of the Base Weather Model function that returns a list of known weather factors for the model.
        if factors is None:
            return list(self.to_si.keys())

        new_factors = []

        for f in factors:
            f_low = f.lower()

            if f_low in self.to_si:
                new_factors.append(f_low)

        return list(set(new_factors))  # Cleanup any duplicate values and return



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from typing import Any

from fastapi import HTTPException


class UnknownSourceException(HTTPException):  # pragma: no cover
    def __init__(self, detail: Any = None):
        self.detail = detail or "unknown data source id"
        self.status_code = 404


class UnknownModelException(HTTPException):  # pragma: no cover
    def __init__(self, detail: Any = None):
        self.detail = detail or "unknown model id"
        self.status_code = 404


class UnknownDataTimeException(HTTPException):
    pass


class UnknownUnitException(HTTPException):
    pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from weather_provider_api.routers.weather.base_models.source import WeatherSourceBase
from weather_provider_api.routers.weather.sources.knmi.models.actuele_waarnemingen import (
    ActueleWaarnemingenModel,
)
from weather_provider_api.routers.weather.sources.knmi.models.actuele_waarnemingen_register import (
    ActueleWaarnemingenRegisterModel,
)
from weather_provider_api.routers.weather.sources.knmi.models.daggegevens import (
    DagGegevensModel,
)
from weather_provider_api.routers.weather.sources.knmi.models.harmonie_arome import (
    HarmonieAromeModel,
)
from weather_provider_api.routers.weather.sources.knmi.models.pluim import PluimModel
from weather_provider_api.routers.weather.sources.knmi.models.uurgegevens import (
    UurgegevensModel,
)


class KNMI(WeatherSourceBase):
    def __init__(self):
        model_instances = [
            UurgegevensModel(),
            DagGegevensModel(),
            HarmonieAromeModel(),
            PluimModel(),
            ActueleWaarnemingenModel(),
            ActueleWaarnemingenRegisterModel(),
        ]

        self.id = "knmi"
        self.name = "Koninklijk Nederlands Meteorologisch Instituut (KNMI)"
        self.url = "https://knmi.nl/"
        self._models = None
        self._async_models = None

        self.setup_models(model_instances)



# !/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0
"""Module that holds the KNMI Downloader class"""

import os
import re
import shutil
import tempfile
from pathlib import Path
from typing import Union

import requests
from loguru import logger


class KNMIDownloader:
    """This class is based on the access methods used for the example downloads for the KNMI Data Platform.

    Due to the rather unusual method of file acquisition, changes were made to make sure only the intended files get
     downloaded and returned.
    """

    API_URL = os.environ.get("KNMI_API_URL", None)
    API_KEY = os.environ.get("KNMI_API_KEY", None)

    def __init__(
        self,
        dataset_name: str,
        dataset_version: str,
        temporary_download_folder: str = None,
        clear_files_from_folder: bool = False,
    ):
        self.dataset_name = dataset_name
        self.dataset_version = dataset_version

        self.temporary_download_folder = self._get_valid_temporary_folder(
            temporary_download_folder, clear_files_from_folder
        )
        logger.info("KNMIDownloader initialized.")
        logger.info(f"Dataset: [{self.dataset_name}] | Version: [{self.dataset_version}]")
        logger.info(f"Storing files in: [{self.temporary_download_folder}]")

    def validate_knmi_download_settings(self):
        if self.API_URL is None:
            raise ValueError("The KNMI Downloader is missing a KNMI API url!")
        if self.API_KEY is None:
            raise ValueError("The KNMI Downloader is missing an access key!")
        logger.info("KNMI download settings validation access test:")
        self.get_all_available_files()
        logger.info("KNMI download settings validated")

    @staticmethod
    def _get_valid_temporary_folder(temporary_download_folder: Union[str, None], clear_files_from_folder: bool):
        """A function that selects the folder for the KNMIDownloader to use to download files.

        Note that this is a temporary folder, and that

        Args:
            temporary_download_folder:  The folder to download to. If None was passed, the default temporary folder
                                         will be used
            clear_files_from_folder:    A boolean holding if the directory should be emptied if it contains files
                                         (True) or if the existence of files should raise an exception (False).

        Returns:
            An existing pathlib Path holding the (empty) folder to use for downloads

        Raises:
            ...:

        """
        if temporary_download_folder is None:
            # if no folder was given, return the OS's default temp dir
            return Path(tempfile.gettempdir())

        possible_download_folder = Path(temporary_download_folder)
        if not possible_download_folder.exists():
            # if the folder doesn't exist yet, we just create it.
            try:
                possible_download_folder.mkdir(parents=False, exist_ok=False)
                return possible_download_folder
            except FileNotFoundError as fnf_error:
                # The parent folder needs to exist
                raise FileNotFoundError(
                    f"The supplied download folder for KNMIDownloader was invalid. "
                    f"At least the parent folder needs to exist for: "
                    f"[{possible_download_folder}]"
                ) from fnf_error
            except FileExistsError as unknown_exception:
                # File didn't exist before, but does now (shouldn't happen)
                raise FileExistsError(
                    f"Inconsistency error when creating the download folder for KNMIDownloader: "
                    f"Initial check stated the target did not exist, but mkdir states it did!"
                    f" Original error: {unknown_exception}"
                ) from unknown_exception

        if not possible_download_folder.is_dir():
            raise FileExistsError(
                "The given target folder name for KNMIDownloader already existed, but as a file! "
                "The target folder is as such not available for use!"
            )

        if next(possible_download_folder.iterdir(), None) is not None and not clear_files_from_folder:
            # Folder is not empty, and we're not allowed to clear it
            raise FileExistsError(
                "Files exist within the target directory for KNMIDownloader, but file deletion is "
                "not allowed. The target folder can as such not be used!"
            )

        if next(possible_download_folder.iterdir(), None) is not None and clear_files_from_folder:
            # Folder is not empty, but we're allowed to clear it.
            try:
                shutil.rmtree(possible_download_folder)  # Clear everything within the folder
                possible_download_folder.mkdir(parents=False, exist_ok=False)
            except Exception as unknown_exception:
                logger.error("An unexpected Exception occurred while clearing the target directory for KNMIDownloader:")
                raise unknown_exception

        return possible_download_folder

    def get_all_available_files(self) -> list:
        """Requests a list with all available files for the configured dataset.

        Returns:
            A list of strings holding the possible files to request for the configured dataset.

        Notes:
            Depending on the dataset configured, this list may hold quite a large number of files.

        """
        logger.info(
            f"Fetching the full list of files available for dataset [{self.dataset_name}], "
            f"version [{self.dataset_version}]."
        )
        files_in_dataset = requests.get(
            f"{self.API_URL}/datasets/{self.dataset_name}/versions/{self.dataset_version}/files",
            headers={"Authorization": self.API_KEY},
            params={"maxKeys": 500},
            timeout=120,  # two minutes of waiting should be far more than enough for file retrieval
        )
        files_in_dataset = files_in_dataset.json()  # Reformat as JSON for proper access

        return files_in_dataset["files"]  # return only the list of file dictionaries

    def download_specific_file(self, filename: str, filesize: int):
        """Downloads a specific file to the set temporary download folder.

        Args:
            filename: The filename of the file to download
            filesize: The size of the file to download (used to validate the file)

        Returns:
            Nothing

        """
        logger.info(f"KNMIDownloader is starting the download of [{filename}], size [{filesize} bytes]")

        if (
            self.temporary_download_folder.joinpath(filename).exists()
            and self.temporary_download_folder.joinpath(filename).stat().st_size == filesize
        ):
            # A file with the correct size and name already exists. There is no need to re-download it.
            logger.info(f"File [{filename}] already exist. Skipping download.")
            return

        file_request = requests.get(
            f"{self.API_URL}/datasets/{self.dataset_name}/versions/{self.dataset_version}/files/{filename}/url",
            headers={"Authorization": self.API_KEY},
            timeout=120,  # 2 minutes to fetch a download url should be more than enough
        )

        temporary_download_url = file_request.json().get("temporaryDownloadUrl")
        logger.debug(f"Temporary download url for [{filename}] set to: {temporary_download_url}")

        self._download_and_save_file_by_name_and_size(temporary_download_url, filesize)

    def _download_and_save_file_by_name_and_size(self, temporary_download_url: str, filesize: int):
        downloaded_file = requests.get(temporary_download_url, timeout=10 * 60)  # 10 minutes should be enough
        file_name_of_download = str(
            re.findall("filename=(.+)", downloaded_file.headers["content-disposition"])[0]
        ).strip('"')

        # Saving the file:
        try:
            downloaded_file.raise_for_status()
            total_bytes_transferred = 0

            with open(self.temporary_download_folder.joinpath(file_name_of_download), "wb") as file_download:
                for chunk in downloaded_file.iter_content(chunk_size=1024 * 1024 * 20):  # 20Mb chunk size
                    if chunk:
                        file_download.write(chunk)
                        total_bytes_transferred += len(chunk)
        finally:
            logger.debug(f"File download complete: [{self.temporary_download_folder.joinpath(file_name_of_download)}]")

        if total_bytes_transferred != filesize:
            raise EOFError(
                f"The downloaded file for [{file_name_of_download}] did not match the expected file size:"
                f" {filesize} bytes"
            )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from datetime import datetime, timedelta
from typing import Optional, Union

import numpy as np
import pandas as pd
from fastapi import HTTPException
from loguru import logger
from pytz import UTC


def parse_datetime(
    datetime_string,
    round_missing_time_up=False,
    round_to_days=False,
    raise_errors=False,
    loc=None,
) -> Optional[datetime]:
    if datetime_string is None:
        return None

    dt = pd.to_datetime(datetime_string, dayfirst=False, errors="coerce")

    if pd.isnull(dt):
        logger.exception("Error while parsing datetime string", input=datetime_string)
        if raise_errors:
            # Note: replace when FastAPI supports Pydantic models to define query parameters
            # (meaning Validators can be used)
            error_msg = {
                "loc": loc,
                "msg": "invalid datetime format",
                "type": "type_error.datetime",
            }
            raise HTTPException(status_code=422, detail=[error_msg])

        dt = None

    if dt is not None and (round_missing_time_up or round_to_days) and time_unknown(dt, datetime_string):
        if round_to_days:
            dt = dt + timedelta(days=1)
        else:
            dt = dt.replace(hour=23, minute=59, second=59)

    if dt is not None:
        dt = np.datetime64(dt).astype(datetime)

    return dt


def time_unknown(dt: datetime, datetime_string: str):  # pragma: no cover
    if dt.hour == 0 and dt.minute == 0 and dt.second == 0 and ":" not in datetime_string:
        return True
    return False


def validate_begin_and_end(
    start: datetime,
    end: datetime,
    data_start: Union[datetime, None] = None,
    data_end: Union[datetime, None] = None,
):
    """Checks the given date parameters and replaces them with default values if they aren't valid.
    The resulting values are then returned.
    """
    start = start.astimezone(UTC) if start else None
    end = end.astimezone(UTC) if end else None
    data_start = data_start.astimezone(UTC) if data_start else None
    data_end = data_end.astimezone(UTC) if data_end else None

    if data_end is None:
        # Assuming predictions fill in this value, the most recent value for the past is before "now".
        data_end = datetime.now(UTC)

    if data_start is not None and data_start > start:
        # If the starting moment lies before what can be requested, put it at the moment from which it can be requested
        start = data_start

    if data_end is not None and data_end < end:
        end = data_end

    if start >= data_end:
        raise HTTPException(
            422,
            f"Invalid [start] value [{start}]: value lies after last available moment for model ({data_end})",
        )
    if data_start is not None and end <= data_start:
        raise HTTPException(
            422,
            f"Invalid [end] value [{end}]: value lies before first available moment for model ({data_start})",
        )

    if end < start:
        raise HTTPException(
            422,
            f"Invalid [start] and [end] values: [end]({end}) lies before [start]({start})",
        )

    return start, end



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Configuration folder"""

import os
import tempfile
from pathlib import Path

from tomli import load

config_file_path = Path(__file__).parent.joinpath("config.toml")
with config_file_path.open(mode="rb") as file_processor:
    # TODO: Check if config.toml overwrite is set in the environment and replace accordingly!!!
    APP_CONFIG = load(file_processor)


# Settings taken from the environment if available.
APP_DEBUGGING = os.environ.get("WPLA_DEBUG", "False").lower() in (
    "true",
    "1",
    "y",
    "yes",
)
APP_DEPLOYED = os.environ.get("WPLA_DEPLOYED", "False").lower() in (
    "true",
    "1",
    "y",
    "yes",
)
APP_STORAGE_FOLDER = os.environ.get("WPLA_STORAGE_FOLDER", f"{tempfile.gettempdir()}/Weather_Repository")
APP_LOG_LEVEL = os.environ.get("WPLA_LOG_LEVEL", APP_CONFIG["logging"]["log_level"]).upper()
APP_SERVER = os.environ.get("WPLA_SERVER_URL", "http://127.0.0.1:8080")  # 127.0.0.1:8080 for debugging is the default



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0
import glob
import os
import re
import sys
import tarfile
import tempfile
from datetime import datetime
from pathlib import Path
from typing import List, Tuple

import cfgrib
import numpy as np
import pandas as pd
import pytz
import xarray as xr
from dateutil.relativedelta import relativedelta
from loguru import logger

from weather_provider_api.routers.weather.repository.repository import (
    RepositoryUpdateResult,
    WeatherRepositoryBase,
)
from weather_provider_api.routers.weather.sources.knmi.client.knmi_downloader import (
    KNMIDownloader,
)
from weather_provider_api.routers.weather.sources.knmi.knmi_factors import arome_factors
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition
from weather_provider_api.routers.weather.utils.grid_helpers import (
    round_coordinates_to_wgs84_grid,
)


class HarmonieAromeRepository(WeatherRepositoryBase):
    """The Weather Repository class for the 'KNMI - Harmonie Arome' dataset"""

    def __init__(self):
        # Pre-work
        super().__init__()

        self.repository_name = "KNMI Harmonie (Arome)"
        logger.debug(f"Initialized {self.repository_name} repository")

        # Repository settings
        self.file_prefix = "AROME"
        self.runtime_limit = 60 * 60 * 3  # 3 hours maximum runtime
        self.permanent_suffixes = ["0000", "0600", "1200", "1800"]
        self.dataset_name = "harmonie_arome_cy43_p1"
        self.dataset_version = "1.0"
        self.file_identifier_length = 13
        self.time_encoding = "hours since 2018-01-01"  # Used to keep values usable for at least the upcoming decennium

        # Verify if cfgrib is properly installed
        try:
            pass
        except RuntimeError as e:
            logger.warning(f"CFGRIB could not properly be initialized: {e}")
            logger.warning(
                "Due to problems with CFGRIB, this repository will only be able to access existing "
                "data. The repository will not be able to update"
            )

    @property
    def repository_sub_folder(self):
        return "AROME"

    def _get_repo_sub_folder(self):
        return self.repository_sub_folder

    @property
    def first_day_of_repo(self):
        first_day_of_repo = datetime.utcnow() - relativedelta(years=3)  # Three years back
        first_day_of_repo = first_day_of_repo.replace(day=1, hour=0, minute=0, second=0, microsecond=0)  # Start of day
        return first_day_of_repo

    @property
    def last_day_of_repo(self):
        last_day_of_repo = datetime.utcnow().replace(minute=0, second=0, microsecond=0)  # Start of today
        return last_day_of_repo

    def update(self):
        """The implementation of the WeatherRepository update function.

        This function gathers any new 'KNMI - Harmonie Arome' files and processes them into the repository. A
         RepositoryUpdateResult object if returned indication success, time-out or failure.

        """
        # Stop if cfgrib wasn't added into the system modules during boot.
        if "cfgrib" not in sys.modules:
            logger.error('CANNOT PERFORM UPDATE: No valid "cfgrib" installation available.')
            logger.info('Please properly install "cfgrib" and restart the system to enable updates.')
            quit()

        # Cleanup the repository
        self.cleanup()

        logger.info(f"KNMI Arome Update - Storage in: {self.repository_folder} ")

        # Configure initial settings for the update
        start_of_update = datetime.utcnow()
        no_of_items_processed = 0
        average_seconds_per_item = 1500  # Assuming 25 minutes of processing time per month for the first item
        forced_end_of_update = start_of_update + relativedelta(seconds=self.runtime_limit)

        logger.info(f"Update of [{self.repository_name}] started")
        logger.info(f"- A forced end time of [{forced_end_of_update}] was set.")

        prediction_to_evaluate = self.get_most_recent_prediction_moment
        download_folder = Path(tempfile.gettempdir()).joinpath(self.dataset_name)
        knmi_downloader = KNMIDownloader(self.dataset_name, self.dataset_version, str(download_folder), True)
        files_in_dataset = {item["filename"]: item["size"] for item in knmi_downloader.get_all_available_files()}

        while prediction_to_evaluate >= self.first_day_of_repo:
            if no_of_items_processed != 0:
                average_seconds_per_item = (datetime.utcnow() - start_of_update).total_seconds() / no_of_items_processed

            if forced_end_of_update < (datetime.utcnow() + relativedelta(seconds=average_seconds_per_item)):
                logger.info(f"- Not enough time left to update before [{forced_end_of_update}]")
                logger.info(f"Update of [{self.repository_name}] ended...")
                return RepositoryUpdateResult.timed_out

            if not self._prediction_already_available(prediction_to_evaluate):
                logger.debug(f"- Gathering the prediction for: {prediction_to_evaluate}")
                file_to_download = (
                    f"HARM43_V1_P1_{prediction_to_evaluate.year}"
                    f"{str(prediction_to_evaluate.month).zfill(2)}"
                    f"{str(prediction_to_evaluate.day).zfill(2)}"
                    f"{str(prediction_to_evaluate.hour).zfill(2)}.tar"
                )

                if file_to_download in files_in_dataset.keys():
                    # File exists
                    knmi_downloader.download_specific_file(file_to_download, files_in_dataset[file_to_download])
                    self._process_downloaded_file(download_folder, file_to_download, prediction_to_evaluate)
                    no_of_items_processed += 1
                else:
                    logger.info(
                        f"The expected file [{file_to_download}] was not found within the KNMI dataset. "
                        f"Moving on to the next file!"
                    )
            else:
                logger.debug(f"The prediction for [{prediction_to_evaluate}] is already stored in the repository.")

            prediction_to_evaluate -= relativedelta(hours=6)

        logger.info(f"Update of [{self.repository_name}] ended...")
        return RepositoryUpdateResult.completed

    def _prediction_already_available(self, prediction_moment: datetime):
        """Function to evaluate if a prediction is already available within the repository."""
        if self.repository_folder.joinpath(
            f"{self.file_prefix}_{prediction_moment.year}{str(prediction_moment.month).zfill(2)}"
            f"{str(prediction_moment.day).zfill(2)}_{str(prediction_moment.hour).zfill(2)}00.nc"
        ).exists():
            return True

        return False

    def _process_downloaded_file(self, download_folder: Path, filename: str, prediction_time: datetime):
        """The function that processes a number of downloaded files into repository files."""
        stage = "Started unpacking files.."
        try:
            self._unpack_downloaded_file(download_folder, filename)
            stage = "Files were unpacked"
            self._convert_unpacked_data_to_netcdf4_files(download_folder, prediction_time)
            stage = "Data was converted to NetCDF4"
            self._fuse_hourly_netcdf4_files(download_folder, prediction_time)
            stage = "NetCDF4 files were properly fused together"
            self._clear_temp_folder(download_folder)
        except Exception as e:
            logger.warning(f"Processing did not get past stage: {stage}")
            logger.warning(f"The downloaded data could not be properly downloaded: {e}")

    @staticmethod
    def _clear_temp_folder(download_folder: Path):
        """A function that cleans up the temporary download folder to prevent issues with partially written files."""
        logger.debug(f"Emptying the download folder: {download_folder}")
        file_filter = f"{download_folder}/*.*" if str(download_folder)[-1] != "/" else f"{download_folder}*.*"
        for existing_file in glob.glob(file_filter):
            try:
                # Try to delete:
                Path(existing_file).unlink()
            except Exception as e:
                logger.info(f"An error occurred while deleting the file [{existing_file}]: {e}")
                logger.warning("There may be issues while updating using a file with an identical name...")

    @staticmethod
    def _unpack_downloaded_file(download_folder: Path, file_name: str):
        """The function that unpacks downloaded files to prediction files."""
        logger.info(f"Unpacking file: {file_name}")
        try:
            tar = tarfile.open(download_folder.joinpath(file_name))
            for member in tar.getmembers():
                if member.isreg():  # Only process files
                    member.name = os.path.basename(member.name)  # Remove the path to setting it to only the filename.
                    tar.extract(member, download_folder)  # Extract the file to the download folder
            tar.close()
        except Exception as e:
            logger.error(f"The tarfile [{file_name}] could not be unpacked!")
            raise e

    def _convert_unpacked_data_to_netcdf4_files(self, download_folder: Path, prediction_time: datetime):
        """This function converts any unpacked data files into NetCDF4 files"""
        try:
            logger.debug("Import of cfgrib was successful")
        except RuntimeError as e:
            logger.error("CFGRIB was not properly installed. Cannot access GRIB files.")
            raise e

        grib_files_available = glob.glob(
            str(download_folder.joinpath(f'HA43_N20_{prediction_time.strftime("%Y%m%d%H")}00_*_GB'))
        )

        for grib_file in grib_files_available:
            logger.debug(f"Processing GRIB file: {grib_file}")
            self._convert_grib_file_to_NetCDF(Path(grib_file))

        logger.info("All Partial datasets were successfully processed.")

    def _convert_grib_file_to_NetCDF(self, grib_file: Path):
        """A function that converts a Harmonie Arome GRIB-file into a NetCDF4 file

        Args:
            grib_file (Path):   The Path to the GRIB file to convert

        Returns:
            Nothing

        """
        prediction_moment, predicted_hour = self._get_times_from_filename(grib_file.name)
        grib_filestream = cfgrib.FileStream(str(grib_file))
        file_dataset = xr.Dataset()

        for item in grib_filestream.items():
            grib_message = item[1]
            # We skip the rotated grid data of the first file in each file-set and only process the regular_ll grids.
            if grib_message["gridType"] == "regular_ll":
                (
                    field_name,
                    message_dataset,
                ) = self._process_grib_message_to_message_dataset(
                    grib_message=grib_message,
                    prediction_moment=prediction_moment,
                    predicted_hour=predicted_hour,
                )
                if message_dataset:
                    if not file_dataset:
                        file_dataset = message_dataset
                    else:
                        file_dataset[field_name] = message_dataset[field_name]

        filename_to_save_to = Path(grib_file.parents[0]).joinpath(
            f"{self.file_prefix}_{prediction_moment.year}{str(prediction_moment.month).zfill(2)}"
            f"{str(prediction_moment.day).zfill(2)}_{str(prediction_moment.hour).zfill(2)}"
            f"00_prediction_for_{str(predicted_hour).zfill(2)}00.nc"
        )

        file_dataset = file_dataset.unstack("coord")
        file_dataset.time.encoding["units"] = "hours since 2018-01-01"

        file_dataset.to_netcdf(path=filename_to_save_to, format="NETCDF4")
        logger.info(f"Saved partial dataset as: {filename_to_save_to}")

    def _process_grib_message_to_message_dataset(
        self,
        grib_message: cfgrib.Message,
        prediction_moment: datetime,
        predicted_hour: int,
    ) -> (str, xr.Dataset):
        """Args:
            grib_message (grib.Message):    A GRIB message holding the data to process
            prediction_moment (datetime):   The datetime moment the data was generated
            predicted_hour (int):           The hour it is predicting

        Returns:
            (str, xarray.Dataset): A string holding the parameter name and a Xarray Dataset holding the data that was
            converted a dataset.

        """
        # Gather the message's level type and the level it was set to
        msg_level_type = "_".join(re.findall("[A-Z][^A-Z]*", grib_message["typeOfLevel"])).lower()
        msg_level = grib_message["level"]

        # Process the supported factors
        parameter_name = grib_message["parameterName"]
        if parameter_name in arome_factors:
            field_name = arome_factors[parameter_name]

            if arome_factors[parameter_name][0] == "_":
                # Add level unit to the name if needed
                field_name = f'{grib_message["stepType"]}{field_name}'
        else:
            field_name = f'unknown_code_{grib_message["parameterName"]}'

        # Add level information
        if msg_level == 0 and msg_level_type == "above_ground":
            field_name = f"surface_{field_name}"
        else:
            field_name = f"{msg_level}m_{msg_level_type}_{field_name}"

        # Prepare data for dataset creation
        field_name = field_name.strip()  # Strip any excess spaces
        lats, lons = self._build_lat_lon_grid(grib_message)  # Get the dimensions of the message
        field_values = np.reshape(grib_message["values"], len(lats) * len(lons))

        data_dict = {field_name: (["time", "coord"], [field_values])}
        predicted_moment = np.datetime64((prediction_moment + relativedelta(hours=predicted_hour))).astype(
            "datetime64[ns]"
        )

        dataset_coords = {
            "time_of_prediction": [prediction_moment],
            "time": [predicted_moment],
            "coord": pd.MultiIndex.from_product([lats, lons], names=["lat", "lon"]),
        }

        message_dataset = xr.Dataset(data_vars=data_dict, coords=dataset_coords)
        message_dataset.time.encoding["units"] = self.time_encoding

        return field_name, message_dataset

    @property
    def get_most_recent_prediction_moment(self):
        """Return the most recent prediction moment, based the current time in the Netherlands

        Returns:
            A datetime holding the most recent available prediction moment

        """
        knmi_lag_time = 5  # There is a known 5-hour calculation lag to take into account

        # Determine the current CET time (the KNMI works from the Netherlands, and this project from UTC)
        time_cet = self.last_day_of_repo.astimezone(pytz.timezone("Europe/Amsterdam"))
        time_cet += time_cet.utcoffset()  # We add the UTC offset to change the actual values to those hours
        time_cet = time_cet.replace(tzinfo=None, microsecond=0)  # Finish formatting

        time_cet -= relativedelta(hours=knmi_lag_time)  # Subtract the KNMI lag
        rounded_hour = (time_cet.hour // 6) * 6  # Get the nearest preceding hours divisible by six
        time_cet = time_cet.replace(hour=rounded_hour, minute=0, second=0)
        return time_cet

    def _fuse_hourly_netcdf4_files(self, download_folder: Path, prediction_moment: datetime):
        """Args:
            download_folder:
            prediction_moment:

        Returns:

        """
        logger.info(f"Starting merge of all files for: {prediction_moment}")
        existing_prediction_files = sorted(glob.glob(str(Path(download_folder).joinpath(f"{self.file_prefix}*.nc"))))

        fused_dataset = None

        for prediction_file in existing_prediction_files:
            logger.debug(f"- Starting merge of: {prediction_file}")
            with xr.open_dataset(prediction_file) as prediction_file_dataset:
                prediction_file_dataset.load()
                if not fused_dataset:
                    fused_dataset = prediction_file_dataset
                else:
                    fused_dataset = xr.merge([fused_dataset, prediction_file_dataset], compat="no_conflicts")

        filename_to_save_to = self.repository_folder.joinpath(
            f"{self.file_prefix}_{prediction_moment.year}{str(prediction_moment.month).zfill(2)}"
            f"{str(prediction_moment.day).zfill(2)}_{str(prediction_moment.hour).zfill(2)}00.nc"
        )

        logger.debug(f"- Saving prediction [{prediction_moment}] to: {filename_to_save_to}")

        encoding = {v: {"zlib": True, "complevel": 4} for v in fused_dataset.variables}
        fused_dataset.to_netcdf(filename_to_save_to, format="NETCDF4", engine="netcdf4", encoding=encoding)

    @staticmethod
    def _build_lat_lon_grid(
        grib_message: cfgrib.Message,
    ) -> Tuple[List[float], List[float]]:
        """This function uses an existing GRIB file to extract the dimensions of the 'regular_ll' grid and format
         those into a list of latitudes and a list of longitudes that together make up the grid.

        Args:
            grib_message:  The GRIB file to use to

        Returns:
            Tuple[List[float], List[float]]: Two lists of float values, the first holding latitudes, and the second
                                              longitudes

        """
        # Try to import cfgrib and raise an error if it isn't properly installed.
        # Set the boundaries
        minimum_latitude = float(grib_message["latitudeOfFirstGridPointInDegrees"])
        maximum_latitude = float(grib_message["latitudeOfLastGridPointInDegrees"])

        minimum_longitude = float(grib_message["longitudeOfFirstGridPointInDegrees"])
        maximum_longitude = float(grib_message["longitudeOfLastGridPointInDegrees"])

        # And the step value needed to get all the inbetween values
        step_for_latitude = int(grib_message["jDirectionIncrement"])
        step_for_longitude = int(grib_message["iDirectionIncrement"])

        # Build the lists
        latitudes = list(
            range(  # We add the step-value to the max-value to include it itself.
                int(minimum_latitude * 1_000),
                int(maximum_latitude * 1_000 + step_for_latitude),
                step_for_latitude,
            )
        )
        longitudes = list(
            range(  # We add the step-value to the max-value to include it itself.
                int(minimum_longitude * 1_000),
                int(maximum_longitude * 1_000 + step_for_longitude),
                step_for_longitude,
            )
        )

        latitudes = [x / 1_000 for x in latitudes]
        longitudes = [y / 1_000 for y in longitudes]

        return latitudes, longitudes

    @staticmethod
    def _get_times_from_filename(filename: str) -> Tuple[datetime, int]:
        """This function extracts the timeframe a file represents from its name and translates that into the datetime
         that the prediction was made, and the exact hour it represents of that prediction.

        Args:
            filename:   The filename to parse

        Returns:
            A datetime and an int value indicating the prediction datetime and the predicted hour respectively

        """
        try:
            prediction_year = int(filename[9:13])
            prediction_month = int(filename[13:15])
            prediction_day = int(filename[15:17])
            prediction_hour = int(filename[17:19])

            predicted_hour = int(filename[-7:-5])
        except Exception as e:
            raise ValueError(f"Could not properly obtain the prediction from the title of the file: {filename} [{e}]")

        prediction_moment = datetime(
            year=prediction_year,
            month=prediction_month,
            day=prediction_day,
            hour=prediction_hour,
        )
        logger.debug(
            f"File [{filename}] was parsed to prediction moment and hour: [{prediction_moment}],[{predicted_hour}]"
        )
        return prediction_moment, predicted_hour

    def _delete_files_outside_of_scope(self):
        """The function that delete all out-of-scope files.

        This means that any file that does no longer exist within scope (either by existing before it, or by existing
         past it) will be removed.

        Returns:
            Nothing.

        Raises:


        """
        counter_until_date_in_filename = len(str(self.repository_folder.joinpath(self.file_prefix))) + 1

        for file_name in glob.glob(f"{self.repository_folder.joinpath(self.file_prefix)}*.nc"):
            file_date = datetime(
                year=int(file_name[counter_until_date_in_filename : counter_until_date_in_filename + 4]),
                month=int(file_name[counter_until_date_in_filename + 4 : counter_until_date_in_filename + 6]),
                day=int(file_name[counter_until_date_in_filename + 6 : counter_until_date_in_filename + 8]),
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ).date()

            if file_date < self.first_day_of_repo.date() or file_date > self.last_day_of_repo.date():
                logger.debug(f"The current scope is [{self.first_day_of_repo.date()}-{self.last_day_of_repo.date()}]")
                logger.debug(f"The file [{file_name}] resolved to [{file_date}], which lies out of scope.")
                logger.debug(f"The file [{file_name}] will therefor be deleted.")

                self._safely_delete_file(file_name)

    def _get_file_list_for_period(self, start: datetime, end: datetime) -> List[Path]:
        """This function creates a list of the files associated with the requested timeframe.

        Args:
            start (datetime):    The start of the timeframe to use
            end (datetime):       The end of the timeframe to use

        Returns:
            List[Path]: A list of the files associated with the given timeframe

        """
        logger.debug(f"Finding files for the timeframe: [({start})-({end})]")
        logger.debug(f"Searching folder: [{self.repository_folder}]")
        counter_until_date_in_filename = len(str(self.repository_folder.joinpath(self.file_prefix))) + 1

        list_of_all_netcdf4_files_in_repo = glob.glob(str(self.repository_folder.joinpath(f"{self.file_prefix}*.nc")))
        list_of_files_in_timeframe = []

        for file_name in list_of_all_netcdf4_files_in_repo:
            file_date = datetime(
                year=int(file_name[counter_until_date_in_filename : counter_until_date_in_filename + 4]),
                month=int(file_name[counter_until_date_in_filename + 4 : counter_until_date_in_filename + 6]),
                day=int(file_name[counter_until_date_in_filename + 6 : counter_until_date_in_filename + 8]),
                hour=0,
                minute=0,
                second=0,
                microsecond=0,
            ).date()

            # Add to the list if within the timeframe
            if start.date() <= file_date <= end.date():
                list_of_files_in_timeframe.append(Path(file_name))

        return list_of_files_in_timeframe

    def get_grid_coordinates(self, coordinates: List[GeoPosition]) -> List[GeoPosition]:
        """Rounds a list of GeoPositions to the resolution set through grid_resolution"""
        return round_coordinates_to_wgs84_grid(coordinates, (0.023, 0.037), (49, 0))



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from weather_provider_api.routers.weather.sources.cds.client.era5sl_repository import (
    ERA5SLRepository,
)


def main() -> None:
    """Simple method wrapper for purging data."""
    era5sl_repo = ERA5SLRepository()
    era5sl_repo.purge_repository()


if __name__ == "__main__":
    main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""CORS support"""

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from loguru import logger

from weather_provider_api.config import APP_CONFIG


def initialize_cors_middleware(app: FastAPI):
    """Initializes the CORS middleware.

    Enables CORS handling for the allowed origins set in the config setting `CORS_ALLOWED_ORIGINS`, a list of strings
     each corresponding to a host. The FastAPI CORSMiddleware is used to enable CORS handling.

    Args:
        app (FastAPI):  The FastAPI application to attach the CORS middleware to.

    Returns:
        Nothing. The application itself is updated.

    """
    origins = APP_CONFIG["components"]["cors_allowed_origins"]
    origins_regex = APP_CONFIG["components"]["cors_allowed_origins_regex"]

    if origins and len(origins) == 0:
        origins = None
    if origins_regex and len(origins_regex) == 0:
        origins_regex = None

    if origins and origins_regex:
        app.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
            allow_origins_regex=origins_regex,
        )
        logger.info(f"Attached CORS middleware enabled for the following origins: {origins}, {origins_regex}")
    elif origins:
        app.add_middleware(
            CORSMiddleware,
            allow_origins=origins,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        logger.info(f"Attached CORS middleware enabled for the following origins: {origins}")
    elif origins_regex:
        app.add_middleware(
            CORSMiddleware,
            allow_origins_regex=origins_regex,
            allow_credentials=True,
            allow_methods=["*"],
            allow_headers=["*"],
        )
        logger.info(f"Attached CORS middleware enabled for the following origins: {origins_regex}")
    else:
        logger.warning("CORS middleware enabled but no allowed origins are set.")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

import json
import math
from collections import defaultdict
from dataclasses import dataclass
from enum import Enum
from typing import Any, List

from fastapi import Query
from pydantic import Field
from starlette.responses import Response as StarletteResponse

from weather_provider_api.core.base_model import BaseModel

"""
    API request models, response models, and examples.
"""


# Note: while this can be done automatically, and for e.g. sources as well,
# I don't want new OpenAPI specs to be created when e.g. a source is added.


class OutputUnit(str, Enum):
    # Valid output unit sets
    si = "si"
    human = "human"
    original = "original"


class ResponseFormat(str, Enum):
    # Valid output file-formats
    netcdf4 = "netcdf4"
    netcdf3 = "netcdf3"
    json = "json"
    json_dataset = "json_dataset"
    csv = "csv"


class WeatherModel(BaseModel):
    id: str = Field(..., description="Model id")
    name: str = Field(..., description="Model name")
    version: str = Field("", description="Model version")
    url: str = Field("", description="Model URL")
    description: str = Field("", description="Model description")
    predictive: bool = Field(..., description="Predictions or measurements")
    async_model: bool = Field(..., description="Whether the model should be called asynchronously")
    time_step_size_minutes: int = Field(..., description="Time between each measurement or prediction")
    num_time_steps: int = Field(..., description="Number of data points in the result set")


class WeatherSource(BaseModel):
    id: str = Field(..., description="Source id")
    name: str = Field(..., description="Source name")
    url: str = Field(None, description="Source URL")
    models: List[WeatherModel] = Field(None, description="Synchronous models")
    async_models: List[WeatherModel] = Field(None, description="Asynchronous models")


@dataclass
class WeatherFormattingRequestQuery:
    units: OutputUnit = Query(OutputUnit.si, description="Unit of weather factors")
    response_format: ResponseFormat = Query(
        ResponseFormat.netcdf4,
        description="Response format (overrides mime-types from Accept HTTP header)",
    )


# Note: I'd love to combine the (almost) duplicate entries below, but the hybrid solutions don't work in FastAPI 0.30.
@dataclass
class WeatherContentRequestQuery:
    # TODO: Change Example value to current date base values
    begin: str = Query(None, description="From date and time", example="2019-01-01 00:00")
    end: str = Query(None, description="To date and time", example="2019-01-31 23:59")
    lat: float = Query(..., description="GPS Latitude or RD x-coordinate", example=52.10)
    lon: float = Query(..., description="GPS Longitude or RD y-coordinate", example=5.18)
    factors: List[str] = Query(None, description="Only return these weather factors (default: all factors)")


@dataclass
class WeatherContentRequestMultiLocationQuery:
    begin: str = Query(None, description="From date and time", example="2019-01-01 00:00")
    end: str = Query(None, description="To date and time", example="2019-01-31 23:59")
    locations: str = Query(
        None,
        description="Locations in either WGS84 (lat,lon) or RD (x,y) format, " "in parentheses, separated by a comma",
        example="(52.1, 5.18), (52.2, 5.22)",
    )
    factors: List[str] = Query(None, description="Only return these weather factors (default: all factors)")


class WeatherContentRequestBody(BaseModel):
    begin: str = Field(None, description="From date and time", example="2019-01-01 00:00")
    end: str = Field(None, description="To date and time", example="2019-01-31 23:59")
    lat: float = Field(..., description="GPS Latitude or RD x-coordinate", example=52.10)
    lon: float = Field(..., description="GPS Longitude or RD y-coordinate", example=5.18)
    factors: List[str] = Field(None, description="Only return these weather factors (default: all factors)")


class ScientificJSONResponse(StarletteResponse):
    media_type = "application/json"

    def render(self, content: Any) -> bytes:
        return json.dumps(
            content,
            ensure_ascii=False,
            allow_nan=True,
            indent=None,
            separators=(",", ":"),
            default=str,
            cls=FloatEncoder,
        ).encode("utf-8")


class FloatEncoder(json.JSONEncoder):
    INFINITY = float("inf")

    def iterencode(self, obj, **kwargs):
        if isinstance(obj, float):
            if math.isnan(obj):  # This checks for NaNs ;-)
                yield "null"
            elif obj == FloatEncoder.INFINITY:
                yield "Infinity"
            elif obj == -FloatEncoder.INFINITY:
                yield "-Infinity"
            else:
                yield format(obj, ".4f")
        elif isinstance(obj, dict):
            last_index = len(obj) - 1
            yield "{"
            i = 0
            for key, value in obj.items():
                yield '"' + key + '": '
                for chunk in FloatEncoder.iterencode(self, value):
                    yield chunk
                if i != last_index:
                    yield ", "
                i += 1
            yield "}"
        elif isinstance(obj, list):
            last_index = len(obj) - 1
            yield "["
            for i, o in enumerate(obj):
                for chunk in FloatEncoder.iterencode(self, o):
                    yield chunk
                if i != last_index:
                    yield ", "
            yield "]"
        else:
            for chunk in json.JSONEncoder.iterencode(self, obj):
                yield chunk


result_mime_types = defaultdict(
    lambda: ResponseFormat.netcdf4,
    {
        "application/netcdf": ResponseFormat.netcdf4,
        "application/netcdf4": ResponseFormat.netcdf4,
        "application/x-netcdf": ResponseFormat.netcdf4,
        "application/netcdf3": ResponseFormat.netcdf3,
        "application/x-netcdf3": ResponseFormat.netcdf3,
        "application/json": ResponseFormat.json,
        "application/json-dataset": ResponseFormat.json_dataset,
        "text/csv": ResponseFormat.csv,
    },
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""Main executable module."""

import uvicorn
from loguru import logger

from weather_provider_api.core.initializers.logging_handler import initialize_logging

# Logging is initialized before the importing of APP_CONFIG, to ensure custom logging for APP_CONFIG initialisation.
initialize_logging()

# Import application configuration settings
from weather_provider_api.config import APP_CONFIG


def launch_api(run_mode: str = "uvicorn", host: str = "127.0.0.1", port: int = 8080):
    """The main method for running this application directly.

    (The Dockerfile uses the WPLA_APPLICATION object in [vbd_memo_api.core.application].)

    Args:
        run_mode (str): The run mode for the application. Accepted values are 'uvicorn' and 'gunicorn'.
        host (str):     The host id to run this application on. Usually 'localhost', '127.0.0.1' or '0.0.0.0' in this
                         context.
        port (str):     The port to broadcast the application at.

    Returns:
        Nothing. Either runs successfully until stopped, or breaks from an Exception.

    Notes:
        As Gunicorn only works within the Linux OS, the 'gunicorn' run_mode setting will not work from any other OS.

    """
    project_title = (
        APP_CONFIG["base"]["full_title"] if APP_CONFIG["base"]["full_title"] else APP_CONFIG["base"]["title"]
    )
    launch_string = f"Launching: {project_title}..."
    logger.info("-" * len(launch_string))
    logger.info(launch_string)
    logger.info("-" * len(launch_string))

    from weather_provider_api.core.application import WPLA_APPLICATION

    # start application based on parameters
    if run_mode.upper() == "UVICORN":
        uvicorn.run(WPLA_APPLICATION, host=host, port=port)
    elif run_mode.upper() == "GUNICORN":
        # Error handling for problems with the Gunicorn application get handled by Gunicorn itself.
        from weather_provider_api.core.gunicorn_application import GunicornApplication

        gunicorn_app = GunicornApplication(options={"bind": f"{host}:{port}"}, fastapi_application=WPLA_APPLICATION)
        gunicorn_app.run()
    else:
        raise ValueError(f"Invalid run-mode selected: {run_mode}")

    shutting_down_string = f"Shutting down: {project_title}"
    logger.info("-" * len(shutting_down_string))
    logger.info(shutting_down_string)
    logger.info("-" * len(shutting_down_string))


# The main function for easy local execution
if __name__ == "__main__":
    # If run from main start using the defaults
    launch_api()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

from weather_provider_api.core.base_model import BaseModel


class ErrorResponseModel(BaseModel):
    detail: str  # error message



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2023 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""(sub)-API Mounting

This module handles the mounting of sub-API's onto a main API.
"""

from fastapi import FastAPI
from loguru import logger

from weather_provider_api.core.initializers.exception_handling import (
    initialize_exception_handler,
)


def mount_api_version(base_application: FastAPI, api_to_mount: FastAPI):
    """The method that mounts a FastAPI object as a child into another.

    Args:
        base_application:   The FastAPI application to be used as the parent application.
        api_to_mount:       The FastAPI application to be used as the child application
    Returns:
        Nothing. The parent object itself is updated after this method.

    Notes:
        This implementation assumes that the child application at least has a [title] and [root_path] set, and that the
         parent application at least has a title set.
    """
    initialize_exception_handler(api_to_mount)  # Hookup the exception handler to the new sub-API
    base_application.mount(api_to_mount.root_path, api_to_mount)
    logger.info(f"Mounted API [{api_to_mount.title}] to [{base_application.title}] at: {api_to_mount.root_path}")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

#  SPDX-FileCopyrightText: 2019-2022 Alliander N.V.
#  SPDX-License-Identifier: MPL-2.0

"""KNMI current weather data fetcher."""

import copy
from typing import List, Optional

import numpy as np
import xarray as xr

from weather_provider_api.routers.weather.base_models.model import WeatherModelBase
from weather_provider_api.routers.weather.sources.knmi.stations import (
    stations_actual,
)
from weather_provider_api.routers.weather.sources.knmi.utils import (
    download_actuele_waarnemingen_weather,
    find_closest_stn_list,
)
from weather_provider_api.routers.weather.utils.geo_position import GeoPosition
from weather_provider_api.routers.weather.utils.pandas_helpers import coords_to_pd_index


class ActueleWaarnemingenModel(WeatherModelBase):
    """A Weather Model that incorporates the "KNMI Actuele Waarnemingen"
    dataset into the Weather Provider API.
    """

    def __init__(self):
        super().__init__()
        self.id = "waarnemingen"
        self.name = "KNMI Actuele Waarnemingen"
        self.version = ""
        self.url = "https://www.knmi.nl/nederland-nu/weer/waarnemingen"
        self.predictive = False
        self.time_step_size_minutes = 10
        self.num_time_steps = 1
        self.description = "Current weather observations. Updated every 10 minutes."
        self.async_model = False

        self.to_si = {
            "weather_description": {
                "name": "weather_description",
                "convert": self.no_conversion,
            },
            "temperature": {"convert": self.celsius_to_kelvin},
            "humidity": {"convert": self.percentage_to_frac},
            "wind_direction": {"convert": self.dutch_wind_direction_to_degrees},
            "wind_speed": {"convert": self.no_conversion},  # m/s
            "visibility": {"convert": self.no_conversion},  # m
            "air_pressure": {"convert": lambda x: x * 100},  # hPa to Pa
        }
        self.to_human = copy.deepcopy(self.to_si)
        self.to_human["temperature"]["convert"] = self.no_conversion  # C

        self.human_to_model_specific = self._create_reverse_lookup(self.to_si)

    def get_weather(
        self,
        coords: List[GeoPosition],
        begin: Optional[np.datetime64],
        end: Optional[np.datetime64],
        weather_factors: List[str] = None,
    ) -> xr.Dataset:
        """The function that gathers and processes the requested Actuele Waarnemingen weather data from the KNMI site
        and returns it as a Xarray Dataset.
        (This model interprets directly from an HTML page, but the information is also available from the data
        platform. Due to it being rather impractically handled, we stick to the site for now.)

        Args:
            coords:             A list of GeoPositions containing the locations the data is requested for.
            begin:              A datetime containing the start of the period to request data for.
            end:                A datetime containing the end of the period to request data for.
            weather_factors:    A list of weather factors to request data for (in string format)

        Returns:
            An Xarray Dataset containing the weather data for the requested period, locations and factors.

        Notes:
            As this model only return the current weather data the 'begin' and 'end' values are not actually used.
        """
        updated_weather_factors = self._request_weather_factors(weather_factors)

        # Download the current weather data
        raw_ds = download_actuele_waarnemingen_weather()

        # Get a list of the relevant STNs and choose the closest STN for each coordinate
        coords_stn, _, _ = find_closest_stn_list(stations_actual, coords)

        # Select the data for the found closest STNs
        ds = raw_ds.sel(STN=coords_stn)

        data_dict = {
            var_name: (["time", "coord"], var.values)
            for var_name, var in ds.data_vars.items()
            if var_name in updated_weather_factors and var_name not in ["lat", "lon"]
        }

        timeline = ds.coords["time"].values

        ds = xr.Dataset(
            data_vars=data_dict,
            coords={"time": timeline, "coord": coords_to_pd_index(coords)},
        )
        ds = ds.unstack("coord")
        return ds

    def is_async(self):  # pragma: no cover
        return self.async_model

    def _request_weather_factors(self, factors: Optional[List[str]]) -> List[str]:
        # Implementation of the Base Weather Model function that returns a list of known weather factors for the model.
        if factors is None:
            return list(self.to_si.keys())

        new_factors = []

        for f in factors:
            f_low = f.lower()
            if f_low in self.to_si:
                new_factors.append(f_low)

        return list(set(new_factors))  # Cleanup any duplicate values and return
